/* hgTracks - Human Genome browser main cgi script. */
#include "common.h"
#include "hCommon.h"
#include "portable.h"
#include "memalloc.h"
#include "obscure.h"
#include "dystring.h"
#include "hash.h"
#include "jksql.h"
#include "memgfx.h"
#include "cheapcgi.h"
#include "htmshell.h"
#include "hdb.h"
#include "psl.h"
#include "agpFrag.h"
#include "agpGap.h"
#include "ctgPos.h"
#include "clonePos.h"
#include "genePred.h"
#include "glDbRep.h"
#include "rmskOut.h"
#include "bed.h"
#include "isochores.h"
#include "simpleRepeat.h"
#include "cpgIsland.h"
#include "cytoBand.h"
#include "gcPercent.h"
#include "genomicDups.h"
#include "mapSts.h"
#include "est3.h"
#include "spaceSaver.h"
#include "wormdna.h"

int maxItemsInFullTrack = 300;

/* These variables persist from one incarnation of this program to the
 * next - living mostly in hidden variables. */
char *chromName;		/* Name of chromosome sequence . */
char *position; 		/* Name of position. */
int winStart;			/* Start of window in sequence. */
int winEnd;			/* End of window in sequence. */
boolean seqReverse;		/* Look at reverse strand. */

boolean withLeftLabels = TRUE;		/* Display left labels? */
boolean withCenterLabels = TRUE;	/* Display center labels? */
boolean withGuidelines = TRUE;		/* Display guidelines? */

boolean calledSelf;			/* True if program called by page it generated. */

char *chromNames[] =
    {
    "chr1",
    "chr2",
    "chr3",
    "chr4",
    "chr5",
    "chr6",
    "chr7",
    "chr8",
    "chr9",
    "chr10",
    "chr11",
    "chr12",
    "chr13",
    "chr14",
    "chr15",
    "chr16",
    "chr17",
    "chr18",
    "chr19",
    "chr20",
    "chr21",
    "chr22",
    "chrX",
    "chrY",
    "chr1_random",
    "chr2_random",
    "chr3_random",
    "chr4_random",
    "chr5_random",
    "chr6_random",
    "chr7_random",
    "chr8_random",
    "chr10_random",
    "chr11_random",
    "chr12_random",
    "chr13_random",
    "chr14_random",
    "chr15_random",
    "chr16_random",
    "chr17_random",
    "chr18_random",
    "chr19_random",
    "chr20_random",
    "chr21_random",
    "chr22_random",
    "chrX_random",
    "chrNA_random",
    };
int chromCount = ArraySize(chromNames);

struct trackLayout
/* This structure controls the basic dimensions of display. */
    {
    MgFont *font;		/* What font to use. */
    int leftLabelWidth;		/* Width of left labels. */
    int trackWidth;		/* Width of tracks. */
    int picWidth;		/* Width of entire picture. */
    } tl;

void initTl()
/* Initialize layout around small font and a picture about 600 pixels
 * wide. */
{
MgFont *font;
font = tl.font = mgSmallFont();
tl.leftLabelWidth = 100;
tl.picWidth = 610;
tl.trackWidth = tl.picWidth - tl.leftLabelWidth;
}

enum trackVisibility 
/* How to look at a track. */
    {
    tvHide=0, 		/* Hide it. */
    tvDense=1,        /* Squish it together. */
    tvFull=2        /* Expand it out. */
    };  

char *tvStrings[] = 
/* User interface strings for above. */
    {
    "hide",
    "dense",
    "full",
    };

/* Other global variables. */
int seqBaseCount;	/* Number of bases in sequence. */
int winBaseCount;	/* Number of bases in window. */

struct trackGroup
/* Structure that displays a group of tracks. */
    {
    struct trackGroup *next;   /* Next on list. */
    char *mapName;             /* Name on image map and for ui buttons. */
    enum trackVisibility visibility; /* How much of this to see if possible. */
    enum trackVisibility limitedVis; /* How much of this actually see. */

    char *longLabel;           /* Long label to put in center. */
    char *shortLabel;          /* Short label to put on side. */

    boolean ignoresColor;      /* True if doesn't use colors below. */
    struct rgbColor color;     /* Main color. */
    Color ixColor;             /* Index of main color. */
    struct rgbColor altColor;  /* Secondary color. */
    Color ixAltColor;

    void (*loadItems)(struct trackGroup *tg);
    /* loadItems loads up items for the chromosome range indicated.  It also usually sets the
     * following variables.  */ 
    void *items;               /* Some type of slList of items. */

    char *(*itemName)(struct trackGroup *tg, void *item);
    /* Return name of one of a member of items above to display on left side. */

    char *(*mapItemName)(struct trackGroup *tg, void *item);
    /* Return name to associate on map. */

    int (*totalHeight)(struct trackGroup *tg, enum trackVisibility vis);
        /* Return total height. Called before and after drawItems. 
         * Must set the following variables. */
    int height;                /* Total height - must be set by above call. */
    int lineHeight;            /* Height per track including border. */
    int heightPer;             /* Height per track minus border. */

    int (*itemHeight)(struct trackGroup *tg, void *item);
    /* Return height of one item. */

    void (*drawItems)(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis);
    /* Draw item list, one per track. */

    void (*itemMap)(struct trackGroup *tg, void *item, int x, int y, int width, int height);
    /* Print image map line on item. */

    void (*freeItems)(struct trackGroup *tg);
    /* Free item list. */

    Color (*itemColor)(struct trackGroup *tg, void *item);
    /* Get color of item (optional). */

    void *customPt;            /* Misc pointer variable unique to group. */
    int customInt;             /* Misc int variable unique to group. */
    int subType;               /* Variable to say what subtype this is for similar groups
                                * to share code. */
    };
struct trackGroup *tGroupList = NULL;  /* List of all tracks. */

static boolean tgLoadNothing(){return TRUE;}
static void tgDrawNothing(){}
static void tgFreeNothing(){}
static char *tgNoName(){return"";}

static int tgFixedItemHeight(struct trackGroup *tg, void *item)
/* Return item height for fixed height track. */
{
return tg->lineHeight;
}

static int tgFixedTotalHeight(struct trackGroup *tg, enum trackVisibility vis)
/* Most fixed height track groups will use this to figure out the height they use. */
{
tg->lineHeight = mgFontLineHeight(tl.font)+1;
tg->heightPer = tg->lineHeight - 1;
switch (vis)
    {
    case tvFull:
	tg->height = slCount(tg->items) * tg->lineHeight;
	break;
    case tvDense:
	tg->height = tg->lineHeight;
	break;
    }
return tg->height;
}


void makeHiddenVar(char *name, char *val)
/* Make hidden variable. */
{
cgiMakeHiddenVar(name, val);	/* Variable set to let us know we called ourselves. */
putchar('\n');
}

void makeHiddenBoolean(char *name, boolean val)
{
if (val)
    makeHiddenVar(name, "1");
}

void saveHiddenVars()
/* Save all the variables we want to continue between states. */
{
char buf[16];
putchar('\n');
makeHiddenVar("old", chromName);    /* Variable set when calling ourselves. */
makeHiddenBoolean("seqReverse", seqReverse);
}

void mapBoxReinvoke(int x, int y, int width, int height, struct trackGroup *toggleGroup, char *chrom,
	int start, int end, char *message)
/* Print out image map rectangle that would invoke this program again.
 * If toggleGroup is non-NULL then toggle that track between full and dense.
 * If chrom is non-null then jump to chrom:start-end. */
{
struct trackGroup *tg;

printf("<AREA SHAPE=RECT COORDS=\"%d,%d,%d,%d\" ", x, y, x+width, y+height);
if (chrom == NULL)
    {
    chrom = chromName;
    start = winStart;
    end = winEnd;
    }
printf("HREF=\"../cgi-bin/hgTracks?seqName=%s&old=%s&winStart=%d&winEnd=%d", 
	chrom, chrom, start, end);
if (withLeftLabels)
    printf("&leftLabels=on");
if (withCenterLabels)
    printf("&centerLabels=on");
if (withGuidelines)
    printf("&guidelines=on");
for (tg = tGroupList; tg != NULL; tg = tg->next)
    {
    int vis = tg->visibility;
    if (tg == toggleGroup)
	{
	if (vis == tvDense)
	    vis = tvFull;
	else if (vis == tvFull)
	    vis = tvDense;
	}
    printf("&%s=%s", tg->mapName, tvStrings[vis]);
    }
if (toggleGroup)
    printf("\" ALT= \"Change between dense and full view of %s track\">\n", toggleGroup->shortLabel);
else
    printf("\" ALT= \"jump to %s\">\n", message);
}


void mapBoxToggleVis(int x, int y, int width, int height, struct trackGroup *curGroup)
/* Print out image map rectangle that would invoke this program again.
 * program with the current track expanded. */
{
mapBoxReinvoke(x, y, width, height, curGroup, NULL, 0, 0, NULL);
}

void mapBoxJumpTo(int x, int y, int width, int height, char *newChrom, int newStart, int newEnd, char *message)
/* Print out image map rectangle that would invoke this program again
 * at a different window. */
{
mapBoxReinvoke(x, y, width, height, NULL, newChrom, newStart, newEnd, message);
}


char *hgcName = "../cgi-bin/hgc";	/* Path to click processing program. */


void mapBoxHc(int x, int y, int width, int height, char *group, char *item, char *statusLine)
/* Print out image map rectangle that would invoke the htc (human track click)
 * program. */
{
char *encodedItem = cgiEncode(item);
printf("<AREA SHAPE=RECT COORDS=\"%d,%d,%d,%d\" ", x, y, x+width, y+height);
printf("HREF=\"%s?g=%s&i=%s&c=%s&l=%d&r=%d\" ", 
    hgcName, group, encodedItem, chromName, winStart, winEnd);
printf("ALT= \"%s\">\n", statusLine); 
freeMem(encodedItem);
}

boolean tableExists(char *table)
/* Return TRUE if a table exists in database. */
{
struct sqlConnection *conn = hAllocConn();
boolean exists = sqlTableExists(conn, table);
hFreeConn(&conn);
return exists;
}

boolean chromTableExists(char *tabSuffix)
/* Return true if chromosome specific version of table exists. */
{
char table[256];
sprintf(table, "%s%s", chromName, tabSuffix);
return tableExists(table);
}

struct simpleFeature
/* Minimal feature - just stores position in browser coordinates. */
    {
    struct simpleFeature *next;
    int start, end;			/* Start/end in browser coordinates. */
    int grayIx;                         /* Level of gray usually. */
    };

struct linkedFeatures
    {
    struct linkedFeatures *next;
    int start, end;			/* Start/end in browser coordinates. */
    int grayIx;				/* Average of components. */
    char name[32];			/* Accession of query seq. */
    int orientation;                    /* Orientation. */
    struct simpleFeature *components;   /* List of component simple features. */
    };


char *linkedFeaturesName(struct trackGroup *tg, void *item)
/* Return name of item. */
{
struct linkedFeatures *lf = item;
return lf->name;
}

void freeLinkedFeatures(struct linkedFeatures **pList)
/* Free up a linked features list. */
{
struct linkedFeatures *lf;
for (lf = *pList; lf != NULL; lf = lf->next)
    slFreeList(&lf->components);
slFreeList(pList);
}

void freeLinkedFeaturesItems(struct trackGroup *tg)
/* Free up linkedFeaturesTrack items. */
{
freeLinkedFeatures((struct linkedFeatures**)(&tg->items));
}

Color shadesOfGray[10];		/* 10 shades of gray from white to black. */
int maxShade = 9;

enum {blackShadeIx=9,whiteShadeIx=0};

Color whiteIndex()
/* Return index of white. */
{
return shadesOfGray[0];
}

Color blackIndex()
/* Return index of black. */
{
return shadesOfGray[maxShade];
}

Color grayIndex()
/* Return index of gray. */
{
return shadesOfGray[(maxShade+1)/2];
}

Color lightGrayIndex()
/* Return index of light gray. */
{
return shadesOfGray[3];
}

void makeGrayShades(struct memGfx *mg)
/* Make eight shades of gray in display. */
{
int i;
int maxShade = ArraySize(shadesOfGray)-1;
for (i=0; i<=maxShade; ++i)
    {
    struct rgbColor rgb;
    int level = 255 - (255*i/maxShade);
    if (level < 0) level = 0;
    rgb.r = rgb.g = rgb.b = level;
    shadesOfGray[i] = mgFindColor(mg, rgb.r, rgb.g, rgb.b);
    }
}

int percentGrayIx(int percent)
/* Return gray shade corresponding to a number from 50 - 100 */
{
int opPercent = 100-percent;
int adjPercent = 100-opPercent-(opPercent>>1);
int level = (adjPercent*maxShade + 50)/100;
if (level < 0) level = 0;
if (level > maxShade) level = maxShade;
return level;
}

int grayInRange(int val, int minVal, int maxVal)
/* Return gray shade corresponding to a number from minVal - maxVal */
{
int range = maxVal - minVal;
int level;
level = ((val-minVal)*maxShade + (range>>1))/range;
if (level <= 0) level = 1;
if (level > maxShade) level = maxShade;
return level;
}


int mrnaGrayIx(int start, int end, int score)
/* Figure out gray level for an RNA block. */
{
int size = end-start;
int halfSize = (size>>1);
int res;
res =  (maxShade * score + halfSize)/size;
return res;
}


static int cmpLfWhiteToBlack(const void *va, const void *vb)
/* Help sort from white to black. */
{
const struct linkedFeatures *a = *((struct linkedFeatures **)va);
const struct linkedFeatures *b = *((struct linkedFeatures **)vb);
return a->grayIx - b->grayIx;
}

static int cmpLfBlackToWhite(const void *va, const void *vb)
/* Help sort from white to black. */
{
const struct linkedFeatures *a = *((struct linkedFeatures **)va);
const struct linkedFeatures *b = *((struct linkedFeatures **)vb);
return b->grayIx - a->grayIx;
}

void sortByGray(struct trackGroup *tg, enum trackVisibility vis)
/* Sort linked features by grayIx. */
{
if (vis == tvDense)
    slSort(&tg->items, cmpLfWhiteToBlack);
else
    slSort(&tg->items, cmpLfBlackToWhite);
}

static void linkedFeaturesDraw(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
/* Draw contig items. */
{
int baseWidth = seqEnd - seqStart;
struct linkedFeatures *lf;
struct simpleFeature *sf;
int y = yOff;
int heightPer = tg->heightPer;
int lineHeight = tg->lineHeight;
int x1,x2;
int midLineOff = heightPer/2;
boolean isFull = (vis == tvFull);
boolean grayScale = tg->ignoresColor;
Color bColor = tg->ixAltColor;
double scale = width/(double)baseWidth;

if (vis == tvDense)
    sortByGray(tg, vis);
for (lf = tg->items; lf != NULL; lf = lf->next)
    {
    int midY = y + midLineOff;
    int compCount = 0;
    int w;
    if (lf->components != NULL)
	{
	x1 = round((double)((int)lf->start-winStart)*scale) + xOff;
	x2 = round((double)((int)lf->end-winStart)*scale) + xOff;
	w = x2-x1;
	if (isFull)
	    {
	    if (grayScale) bColor =  shadesOfGray[(lf->grayIx>>1)];
	    mgBarbedHorizontalLine(mg, x1, midY, x2-x1, 2, 5, 
		    lf->orientation, bColor);
	    }
	if (grayScale) color =  shadesOfGray[lf->grayIx];
	mgDrawBox(mg, x1, midY, w, 1, color);
	}
    for (sf = lf->components; sf != NULL; sf = sf->next)
	{
	x1 = round((double)((int)sf->start-winStart)*scale) + xOff;
	x2 = round((double)((int)sf->end-winStart)*scale) + xOff;
	w = x2-x1;
	if (w < 1)
	    w = 1;
	if (grayScale) color =  shadesOfGray[lf->grayIx];
	mgDrawBox(mg, x1, y, w, heightPer, color);
	++compCount;
	}
    if (isFull)
	y += lineHeight;
    }
}

void incRange(UBYTE *start, int size)
/* Add one to range of bytes, taking care to not overflow. */
{
int i;
UBYTE b;
for (i=0; i<size; ++i)
    {
    b = start[i];
    if (b < 254)
	start[i] = b+2;
    }
}

void resampleBytes(UBYTE *s, int sct, UBYTE *d, int dct)
/* Shrink or stretch an line of bytes. */
{
#define WHOLESCALE 256
if (sct > dct)	/* going to do some averaging */
	{
	int i;
	int j, jend, lj;
	long lasts, ldiv;
	long acc, div;
	long t1,t2;

	ldiv = WHOLESCALE;
	lasts = s[0];
	lj = 0;
	for (i=0; i<dct; i++)
		{
		acc = lasts*ldiv;
		div = ldiv;
		t1 = (i+1)*(long)sct;
		jend = t1/dct;
		for (j = lj+1; j<jend; j++)
			{
			acc += s[j]*WHOLESCALE;
			div += WHOLESCALE;
			}
		t2 = t1 - jend*(long)dct;
		lj = jend;
		lasts = s[lj];
		if (t2 == 0)
			{
			ldiv = WHOLESCALE;
			}
		else
			{
			ldiv = WHOLESCALE*t2/dct;
			div += ldiv;
			acc += lasts*ldiv;
			ldiv = WHOLESCALE-ldiv;
			}
		*d++ = acc/div;
		}
	}
else if (dct == sct)	/* they's the same */
	{
	while (--dct >= 0)
		*d++ = *s++;
	}
else if (sct == 1)
	{
	while (--dct >= 0)
		*d++ = *s;
	}
else/* going to do some interpolation */
	{
	int i;
	long t1;
	long p1;
	long err;
	int dct2;

	dct -= 1;
	sct -= 1;
	dct2 = dct/2;
	t1 = 0;
	for (i=0; i<=dct; i++)
		{
		p1 = t1/dct;
		err =  t1 - p1*dct;
		if (err == 0)
			*d++ = s[p1];
		else
			*d++ = (s[p1]*(dct-err)+s[p1+1]*err+dct2)/dct;
		t1 += sct;
		}
	}
}

void grayThreshold(UBYTE *pt, int count)
/* Convert from 0-4 representation to gray scale rep. */
{
UBYTE b;
int i;

for (i=0; i<count; ++i)
    {
    b = pt[i];
    if (b == 0)
	pt[i] = shadesOfGray[0];
    else if (b == 1)
	pt[i] = shadesOfGray[2];
    else if (b == 2)
	pt[i] = shadesOfGray[4];
    else if (b == 3)
	pt[i] = shadesOfGray[6];
    else if (b >= 4)
	pt[i] = shadesOfGray[9];
    }
}


static void linkedFeaturesDrawAverage(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
/* Draw dense clone items. */
{
int baseWidth = seqEnd - seqStart;
UBYTE *useCounts;
int i;
int lineHeight = mgFontLineHeight(font);
struct simpleFeature *sf;
struct linkedFeatures *lf;
int x1, x2, w;

AllocArray(useCounts, width);
memset(useCounts, 0, width * sizeof(useCounts[0]));
for (lf = tg->items; lf != NULL; lf = lf->next)
    {
    for (sf = lf->components; sf != NULL; sf = sf->next)
	{
	x1 = roundingScale(sf->start-winStart, width, baseWidth);
	if (x1 < 0)
	   x1 = 0;
	x2 = roundingScale(sf->end-winStart, width, baseWidth);
	if (x2 >= width)
	   x2 = width-1;
	w = x2-x1;
	if (w >= 0)
	    {
	    if (w == 0)
	        w = 1;
	    incRange(useCounts+x1, w); 
	    }
	}
    }
grayThreshold(useCounts, width);
for (i=0; i<lineHeight; ++i)
    mgPutSegZeroClear(mg, xOff, yOff+i, width, useCounts);
freeMem(useCounts);
}

static void linkedFeaturesAverageDense(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
/* Draw contig items. */
{
if (vis == tvFull)
    {
    linkedFeaturesDraw(tg, seqStart, seqEnd, mg, xOff, yOff, width, font, color, vis);
    }
else if (vis == tvDense)
    {
    linkedFeaturesDrawAverage(tg, seqStart, seqEnd, mg, xOff, yOff, width, font, color, vis);
    }
}

int lfCalcGrayIx(struct linkedFeatures *lf)
/* Calculate gray level from components. */
{
struct simpleFeature *sf;
int count = 0;
int total = 0;

for (sf = lf->components; sf != NULL; sf = sf->next)
    {
    ++count;
    total += sf->grayIx;
    }
if (count == 0)
    return whiteShadeIx;
return (total+(count>>1))/count;
}

void finishLf(struct linkedFeatures *lf)
/* Calculate beginning and end of lf from components, etc. */
{
struct simpleFeature *sf;

slReverse(&lf->components);
if ((sf = lf->components) != NULL)
    {
    int start = sf->start;
    int end = sf->end;

    for (sf = sf->next; sf != NULL; sf = sf->next)
	{
	if (sf->start < start)
	    start = sf->start;
	if (sf->end > end)
	    end = sf->end;
	}
    lf->start = start;
    lf->end = end;
    }
lf->grayIx = lfCalcGrayIx(lf);
}

struct trackGroup *linkedFeaturesTg()
/* Return generic track group for linked features. */
{
struct trackGroup *tg = NULL;

AllocVar(tg);
tg->freeItems = freeLinkedFeaturesItems;
tg->drawItems = linkedFeaturesDraw;
tg->ignoresColor = TRUE;
tg->itemName = linkedFeaturesName;
tg->mapItemName = linkedFeaturesName;
tg->totalHeight = tgFixedTotalHeight;
tg->itemHeight = tgFixedItemHeight;
return tg;
}

int orientFromChar(char c)
/* Return 1 or -1 in place of + or - */
{
if (c == '-')
    return -1;
else
    return 1;
}

char charFromOrient(int orient)
/* Return + or - in place of 1 or -1 */
{
if (orient < 0)
    return '-';
else
    return '+';
}

struct linkedFeatures *lfFromPslsInRange(char *table, int start, int end)
/* Return linked features from range of table. */
{
char query[256];
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr = NULL;
char **row;
struct linkedFeatures *lfList = NULL, *lf;

sprintf(query, "select * from %s where tStart<%u and tEnd>%u",
    table, winEnd, winStart);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    struct simpleFeature *sfList = NULL, *sf;
    struct psl *psl = pslLoad(row);
    unsigned *starts = psl->tStarts;
    unsigned *sizes = psl->blockSizes;
    int i, blockCount = psl->blockCount;
    int grayIx = mrnaGrayIx(psl->qStart, psl->qEnd, psl->match - psl->misMatch + psl->repMatch - psl->qNumInsert);

    AllocVar(lf);
    lf->grayIx = grayIx;
    strncpy(lf->name, psl->qName, sizeof(lf->name));
    lf->orientation = orientFromChar(psl->strand[0]);
    for (i=0; i<blockCount; ++i)
	{
	AllocVar(sf);
	sf->start = sf->end = starts[i];
	sf->end += sizes[i];
	sf->grayIx = grayIx;
	slAddHead(&sfList, sf);
	}
    slReverse(&sfList);
    lf->components = sfList;
    finishLf(lf);
    slAddHead(&lfList, lf);
    pslFree(&psl);
    }
slReverse(&lfList);
sqlFreeResult(&sr);
hFreeConn(&conn);
return lfList;
}


void mrnaItemMap(struct trackGroup *tg, void *item, int x, int y, int width, int height)
/* Print image map line on one mRNA or EST item. */
{
struct linkedFeatures *lf = item;
char *id = tg->mapName;
char *shortName = cgiEncode(tg->mapItemName(tg, item));
char *statusLine = cgiEncode(tg->itemName(tg, item));

printf("<AREA SHAPE=RECT COORDS=\"%d,%d,%d,%d\" ", x, y, x+width, y+height);
printf("HREF=\"../cgi-bin/hgc?g=%s&i=%s&c=%s&l=%d&r=%d&o=%d\" ", 
    id, shortName, chromName, winStart, winEnd, lf->start);
printf("ALT= \"%s\" TITLE=\"%s\">\n", statusLine, statusLine); 
freeMem(shortName);
freeMem(statusLine);
}

void loadMrnaAli(struct trackGroup *tg)
/* Load up rnas from table into trackGroup items. */
{
char table[64];
sprintf(table, "%s_mrna", chromName);
tg->items = lfFromPslsInRange(table, winStart, winEnd);
}

struct trackGroup *fullMrnaTg()
/* Make track group of full length mRNAs. */
{
struct trackGroup *tg = linkedFeaturesTg();
tg->mapName = "hgMrna";
tg->visibility = tvFull;
tg->longLabel = "Full Length mRNAs";
tg->shortLabel = "Full mRNAs";
tg->loadItems = loadMrnaAli;
tg->itemMap = mrnaItemMap;
return tg;
}

void loadEstAli(struct trackGroup *tg)
/* Load up rnas from table into trackGroup items. */
{
char table[64];
sprintf(table, "%s_est", chromName);
tg->items = lfFromPslsInRange(table, winStart, winEnd);
}

struct trackGroup *estTg()
/* Make track group of full length mRNAs. */
{
struct trackGroup *tg = linkedFeaturesTg();
tg->mapName = "hgEst";
tg->visibility = tvDense;
tg->longLabel = "Human ESTs";
tg->shortLabel = "Human ESTs";
tg->loadItems = loadEstAli;
tg->drawItems = linkedFeaturesAverageDense;
tg->itemMap = mrnaItemMap;
return tg;
}

struct linkedFeatures *lfFromGenePredInRange(char *table, 
	char *chrom, int start, int end)
/* Return linked features from range of a gene prediction table. */
{
char query[256];
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr = NULL;
char **row;
struct linkedFeatures *lfList = NULL, *lf;
int grayIx = maxShade;

sprintf(query, "select * from %s where chrom='%s' and txStart<%u and txEnd>%u",
    table, chrom, winEnd, winStart);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    struct simpleFeature *sfList = NULL, *sf;
    struct genePred *gp = genePredLoad(row);
    unsigned *starts = gp->exonStarts;
    unsigned *ends = gp->exonEnds;
    int i, blockCount = gp->exonCount;

    AllocVar(lf);
    lf->grayIx = grayIx;
    strncpy(lf->name, gp->name, sizeof(lf->name));
    lf->orientation = orientFromChar(gp->strand[0]);
    for (i=0; i<blockCount; ++i)
	{
	AllocVar(sf);
	sf->start = starts[i];
	sf->end = ends[i];
	sf->grayIx = grayIx;
	slAddHead(&sfList, sf);
	}
    slReverse(&sfList);
    lf->components = sfList;
    finishLf(lf);
    slAddHead(&lfList, lf);
    genePredFree(&gp);
    }
slReverse(&lfList);
sqlFreeResult(&sr);
hFreeConn(&conn);
return lfList;
}

void abbr(char *s, char *fluff)
/* Cut out fluff from s. */
{
int len;
s = strstr(s, fluff);
if (s != NULL)
   {
   len = strlen(fluff);
   strcpy(s, s+len);
   }
}

char *genieName(struct trackGroup *tg, void *item)
/* Return full genie name. */
{
struct linkedFeatures *lf = item;
char *full = lf->name;
static char abbrev[32];

strncpy(abbrev, full, sizeof(abbrev));
abbr(abbrev, "00000");
abbr(abbrev, "0000");
abbr(abbrev, "000");
abbr(abbrev, "ctg");
return abbrev;
}

char *genieMapName(struct trackGroup *tg, void *item)
/* Return abbreviated genie name. */
{
struct linkedFeatures *lf = item;
return lf->name;
}


void loadGenieAlt(struct trackGroup *tg)
/* Load up Genie alt genes. */
{
tg->items = lfFromGenePredInRange("genieAlt", chromName, winStart, winEnd);
}

struct trackGroup *genieAltTg()
/* Make track group of full length mRNAs. */
{
struct trackGroup *tg = linkedFeaturesTg();
tg->mapName = "genieAlt";
tg->visibility = tvHide;
tg->longLabel = "Genie Predictions Using mRNA/ESTs";
tg->shortLabel = "Genie";
tg->loadItems = loadGenieAlt;
tg->ignoresColor = FALSE;
tg->color.r = 125;
tg->color.g = 0;
tg->color.b = 150;
tg->altColor.r = (125+255)/2;
tg->altColor.g = (0+255)/2;
tg->altColor.b = (150+255)/2;
tg->itemName = genieName;
tg->mapItemName = genieMapName;
return tg;
}

void loadGenieKnown(struct trackGroup *tg)
/* Load up Genie known genes. */
{
tg->items = lfFromGenePredInRange("genieKnown", chromName, winStart, winEnd);
}

struct trackGroup *genieKnownTg()
/* Make track group of known genes. */
{
struct trackGroup *tg = linkedFeaturesTg();
tg->mapName = "genieKnown";
tg->visibility = tvFull;
tg->longLabel = "Known Genes (Processed Through Genie)";
tg->shortLabel = "Known Genes";
tg->loadItems = loadGenieKnown;
tg->ignoresColor = FALSE;
tg->color.r = 20;
tg->color.g = 20;
tg->color.b = 170;
tg->altColor.r = (20+255)/2;
tg->altColor.g = (20+255)/2;
tg->altColor.b = (170+255)/2;
tg->itemName = genieName;
tg->mapItemName = genieMapName;
return tg;
}

void loadEnsGene(struct trackGroup *tg)
/* Load up Ensembl genes. */
{
tg->items = lfFromGenePredInRange("ensGene", chromName, winStart, winEnd);
}

char *ensGeneName(struct trackGroup *tg, void *item)
/* Return abbreviated ensemble gene name. */
{
struct linkedFeatures *lf = item;
char *full = lf->name;
static char abbrev[32];

strncpy(abbrev, full, sizeof(abbrev));
abbr(abbrev, "SEPT20T.");
return abbrev;
}

char *ensGeneMapName(struct trackGroup *tg, void *item)
/* Return full ensemble gene name. */
{
struct linkedFeatures *lf = item;
return lf->name;
}

struct trackGroup *ensemblGeneTg()
/* Make track group of genie predictions. */
{
struct trackGroup *tg = linkedFeaturesTg();
tg->mapName = "hgEnsGene";
tg->visibility = tvHide;
tg->longLabel = "Ensembl Gene Predictions";
tg->shortLabel = "Ensembl Genes";
tg->loadItems = loadEnsGene;
tg->ignoresColor = FALSE;
tg->color.r = 150;
tg->color.g = 0;
tg->color.b = 0;
tg->altColor.r = (150+255)/2;
tg->altColor.g = (0+255)/2;
tg->altColor.b = (0+255)/2;
tg->itemName = ensGeneName;
tg->mapItemName = ensGeneMapName;
return tg;
}


void goldLoad(struct trackGroup *tg)
/* Load up golden path from database table to trackGroup items. */
{
char table[64];
char query[256];
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr = NULL;
char **row;
struct agpFrag *fragList = NULL, *frag;
struct agpGap *gapList = NULL, *gap;

/* Get the frags and load into tg->items. */
sprintf(table, "%s_gold", chromName);
sprintf(query, "select * from %s where chromStart<%u and chromEnd>%u",
    table, winEnd, winStart);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    frag = agpFragLoad(row);
    slAddHead(&fragList, frag);
    }
slReverse(&fragList);
sqlFreeResult(&sr);
tg->items = fragList;

/* Get the gaps into tg->customPt. */
sprintf(table, "%s_gap", chromName);
sprintf(query, "select * from %s where chromStart<%u and chromEnd>%u",
    table, winEnd, winStart);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    gap = agpGapLoad(row);
    slAddHead(&gapList, gap);
    }
slReverse(&gapList);
sqlFreeResult(&sr);
tg->customPt = gapList;
hFreeConn(&conn);
}

void goldFree(struct trackGroup *tg)
/* Free up goldTrackGroup items. */
{
agpFragFreeList((struct agpFrag**)&tg->items);
agpGapFreeList((struct agpGap**)&tg->customPt);
}

char *goldName(struct trackGroup *tg, void *item)
/* Return name of gold track item. */
{
struct agpFrag *frag = item;
return frag->frag;
}

static void goldDraw(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
/* Draw golden path items. */
{
int baseWidth = seqEnd - seqStart;
struct agpFrag *frag;
struct agpGap *gap;
int y = yOff;
int heightPer = tg->heightPer;
int lineHeight = tg->lineHeight;
int x1,x2,w;
int midLineOff = heightPer/2;
boolean isFull = (vis == tvFull);
Color brown = color;
Color gold = tg->ixAltColor;
Color col;
int ix = 0;

/* Draw gaps if any. */
if (!isFull)
    {
    int midY = y + midLineOff;
    for (gap = tg->customPt; gap != NULL; gap = gap->next)
	{
	if (!sameWord(gap->bridge, "no"))
	    {
	    x1 = roundingScale(gap->chromStart-winStart, width, baseWidth)+xOff;
	    x2 = roundingScale(gap->chromEnd-winStart, width, baseWidth)+xOff;
	    w = x2-x1;
	    mgDrawBox(mg, x1, midY, w, 1, brown);
	    }
	}
    }

for (frag = tg->items; frag != NULL; frag = frag->next)
    {
    x1 = roundingScale(frag->chromStart-winStart, width, baseWidth)+xOff;
    x2 = roundingScale(frag->chromEnd-winStart, width, baseWidth)+xOff;
    w = x2-x1;
    if (isFull)
	color = gold;
    else
	color =  ((ix&1) ? gold : brown);
    if (w < 1)
	w = 1;
    mgDrawBox(mg, x1, y, w, heightPer, color);
    if (isFull)
	y += lineHeight;
    else if (baseWidth < 10000000)
	{
	char status[256];
	sprintf(status, "%s:%d-%d %s %s:%d-%d", 
	    frag->frag, frag->fragStart, frag->fragEnd,
	    frag->strand,
	    frag->chrom, frag->chromStart, frag->chromEnd);

	mapBoxHc(x1,y,w,heightPer, tg->mapName, 
	    frag->frag, status);
	}
    ++ix;
    }
}

struct trackGroup *goldTrackGroup()
/* Make track group for golden path */
{
struct trackGroup *tg;

AllocVar(tg);
tg->mapName = "hgGold";
tg->visibility = tvHide;
tg->longLabel = "Assembly from Fragments";
tg->shortLabel = "Assembly";
tg->loadItems = goldLoad;
tg->freeItems = goldFree;
tg->drawItems = goldDraw;
tg->color.r = 150;
tg->color.g = 100;
tg->color.b = 30;
tg->altColor.r = 230;
tg->altColor.g = 170;
tg->altColor.b = 40;
tg->itemName = goldName;
tg->mapItemName = goldName;
tg->totalHeight = tgFixedTotalHeight;
tg->itemHeight = tgFixedItemHeight;
return tg;
}

/* Repeat items.  Since there are so many of these, to avoid 
 * memory problems we don't query the database and store the results
 * during repeatLoad, but rather query the database during the
 * actual drawing. */

struct repeatItem
/* A repeat track item. */
    {
    struct repeatItem *next;
    char *class;
    char *className;
    int yOffset;
    };

static struct repeatItem *otherRepeatItem = NULL;
static char *repeatClassNames[] =  {
    "SINE", "LINE", "LTR", "DNA", "Simple", "Low Complexity", "Satellite", "tRNA", "Other",
};
static char *repeatClasses[] = {
    "SINE", "LINE", "LTR", "DNA", "Simple_repeat", "Low_complexity", "Satellite", "tRNA", "Other",
};

struct repeatItem *makeRepeatItems()
/* Make the stereotypical repeat masker tracks. */
{
struct repeatItem *ri, *riList = NULL;
int i;
int numClasses = ArraySize(repeatClasses);
for (i=0; i<numClasses; ++i)
    {
    AllocVar(ri);
    ri->class = repeatClasses[i];
    ri->className = repeatClassNames[i];
    slAddHead(&riList, ri);
    }
otherRepeatItem = riList;
slReverse(&riList);
return riList;
}

void repeatLoad(struct trackGroup *tg)
/* Load up repeat tracks.  (Will query database during drawing for a change.) */
{
tg->items = makeRepeatItems();
}

void repeatFree(struct trackGroup *tg)
/* Free up goldTrackGroup items. */
{
slFreeList(&tg->items);
}

char *repeatName(struct trackGroup *tg, void *item)
/* Return name of repeat item track. */
{
struct repeatItem *ri = item;
return ri->className;
}

static void repeatDraw(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
{
int baseWidth = seqEnd - seqStart;
struct repeatItem *ri;
int y = yOff;
int heightPer = tg->heightPer;
int lineHeight = tg->lineHeight;
int x1,x2,w;
boolean isFull = (vis == tvFull);
Color col;
int ix = 0;
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr = NULL;
char **row;
char query[256];

if (isFull)
    {
    /* Do gray scale representation spread out among tracks. */
    struct hash *hash = newHash(6);
    struct rmskOut ro;
    int percId;
    int grayLevel;
    char statusLine[128];

    for (ri = tg->items; ri != NULL; ri = ri->next)
        {
	ri->yOffset = y;
	y += lineHeight;
	hashAdd(hash, ri->class, ri);
	}
    sprintf(query, "select * from %s_rmsk where genoStart<%u and genoEnd>%u",
	chromName, winEnd, winStart);
    sr = sqlGetResult(conn, query);
    while ((row = sqlNextRow(sr)) != NULL)
        {
	rmskOutStaticLoad(row, &ro);
	ri = hashFindVal(hash, ro.repClass);
	if (ri == NULL)
	   ri = otherRepeatItem;
	percId = 1000 - ro.milliDiv - ro.milliDel - ro.milliIns;
	grayLevel = grayInRange(percId, 500, 1000);
	col = shadesOfGray[grayLevel];
	x1 = roundingScale(ro.genoStart-winStart, width, baseWidth)+xOff;
	x2 = roundingScale(ro.genoEnd-winStart, width, baseWidth)+xOff;
	w = x2-x1;
	if (w <= 0)
	    w = 1;
	mgDrawBox(mg, x1, ri->yOffset, w, heightPer, col);
	if (baseWidth <= 100000)
	    {
	    if (ri == otherRepeatItem)
		{
		sprintf(statusLine, "Repeat %s, family %s, class %s",
		    ro.repName, ro.repFamily, ro.repClass);
		}
	    else
		{
		sprintf(statusLine, "Repeat %s, family %s",
		    ro.repName, ro.repFamily);
		}
	    printf("<AREA SHAPE=RECT COORDS=\"%d,%d,%d,%d\" ", x1, ri->yOffset, x1+w, ri->yOffset+heightPer);
	    printf("HREF=\"../cgi-bin/hgc?g=%s&i=%s&c=%s&l=%d&r=%d&o=%d\" ", 
		tg->mapName, ro.repName, chromName, winStart, winEnd, ro.genoStart);
	    printf("ALT= \"%s\" TITLE=\"%s\">\n", statusLine, statusLine); 
	    }
	}
    freeHash(&hash);
    }
else
    {
    /* Do black and white on single track. */
    sprintf(query, "select genoStart,genoEnd from %s_rmsk where genoStart<%u and genoEnd>%u",
	chromName, winEnd, winStart);
    sr = sqlGetResult(conn, query);
    while ((row = sqlNextRow(sr)) != NULL)
        {
	int start = sqlUnsigned(row[0]);
	int end = sqlUnsigned(row[1]);
	x1 = roundingScale(start-winStart, width, baseWidth)+xOff;
	x2 = roundingScale(end-winStart, width, baseWidth)+xOff;
	w = x2-x1;
	if (w <= 0)
	    w = 1;
	mgDrawBox(mg, x1, yOff, w, heightPer, MG_BLACK);
	}
    }
sqlFreeResult(&sr);
hFreeConn(&conn);
}

struct trackGroup *repeatTg()
/* Make track group for repeats. */
{
struct trackGroup *tg;

AllocVar(tg);
tg->mapName = "hgRepeat";
tg->visibility = tvDense;
tg->longLabel = "Repeating Elements by RepeatMasker";
tg->shortLabel = "RepeatMasker";
tg->loadItems = repeatLoad;
tg->freeItems = repeatFree;
tg->drawItems = repeatDraw;
tg->ignoresColor = TRUE;
tg->itemName = repeatName;
tg->mapItemName = repeatName;
tg->totalHeight = tgFixedTotalHeight;
tg->itemHeight = tgFixedItemHeight;
return tg;
}

typedef struct slList *(*ItemLoader)(char **row);

void bedLoadQuery(struct trackGroup *tg, char *query, ItemLoader loader)
/* Load up bed items from a query. */
{
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr = NULL;
char **row;
struct slList *itemList = NULL, *item;

/* Get the frags and load into tg->items. */
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    item = loader(row);
    slAddHead(&itemList, item);
    }
slReverse(&itemList);
sqlFreeResult(&sr);
tg->items = itemList;
hFreeConn(&conn);
}

void bedLoadChrom(struct trackGroup *tg, char *table, ItemLoader loader)
/* Generic tg->item loader from chromosome table. */
{
char query[256];
sprintf(query, "select * from %s_%s where chromStart<%u and chromEnd>%u",
    chromName, table, winEnd, winStart);
bedLoadQuery(tg, query, loader);
}

void bedLoad(struct trackGroup *tg, char *table, ItemLoader loader)
/* Generic tg->item loader. */
{
char query[256];
sprintf(query, "select * from %s where chrom = '%s' and chromStart<%u and chromEnd>%u",
    table, chromName, winEnd, winStart);
bedLoadQuery(tg, query, loader);
}

static void bedDrawSimple(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
/* Draw simple Bed items. */
{
int baseWidth = seqEnd - seqStart;
struct bed *item;
int y = yOff;
int heightPer = tg->heightPer;
int lineHeight = tg->lineHeight;
int x1,x2,w;
boolean isFull = (vis == tvFull);
double scale = width/(double)baseWidth;
int invPpt;
boolean grayScale = (tg->ignoresColor);

for (item = tg->items; item != NULL; item = item->next)
    {
    x1 = round((double)((int)item->chromStart-winStart)*scale) + xOff;
    x2 = round((double)((int)item->chromEnd-winStart)*scale) + xOff;
    w = x2-x1;
    if (tg->itemColor != NULL)
        color = tg->itemColor(tg, item);
    else
	{
	if (grayScale)
	    color = shadesOfGray[grayInRange(item->score, 0, 1000)];
	}
    if (w < 1)
	w = 1;
    if (color)
	mgDrawBox(mg, x1, y, w, heightPer, color);
    if (isFull)
	{
	char *name = tg->mapItemName(tg, item);
	char *cgiName = cgiEncode(tg->mapItemName(tg, item));
	printf("<AREA SHAPE=RECT COORDS=\"%d,%d,%d,%d\" ", 
		x1, y, x1+w, y+heightPer);
	printf("HREF=\"../cgi-bin/hgc?g=%s&i=%s&c=%s&l=%d&r=%d&o=%d\" ", 
	    tg->mapName, cgiName, 
	    chromName, winStart, winEnd, item->chromStart);
	printf("ALT= \"%s\" TITLE=\"%s\">\n", name, name); 
	freeMem(cgiName);
	y += lineHeight;
	}
    }
}

char *bedName(struct trackGroup *tg, void *item)
/* Return name of bed track item. */
{
struct bed *bed = item;
return bed->name;
}


void isochoreLoad(struct trackGroup *tg)
/* Load up isochores from database table to trackGroup items. */
{
bedLoad(tg, "isochores", (ItemLoader)isochoresLoad);
}

void isochoreFree(struct trackGroup *tg)
/* Free up isochore items. */
{
isochoresFreeList((struct isochores**)&tg->items);
}

char *isochoreName(struct trackGroup *tg, void *item)
/* Return name of gold track item. */
{
struct isochores *iso = item;
static char buf[64];
sprintf(buf, "%3.1f%% GC", 0.1*iso->gcPpt);
return buf;
}

static void isochoreDraw(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
/* Draw isochore items. */
{
int baseWidth = seqEnd - seqStart;
struct isochores *item;
int y = yOff;
int heightPer = tg->heightPer;
int lineHeight = tg->lineHeight;
int x1,x2,w;
boolean isFull = (vis == tvFull);
double scale = width/(double)baseWidth;
int invPpt;

for (item = tg->items; item != NULL; item = item->next)
    {
    x1 = round((double)((int)item->chromStart-winStart)*scale) + xOff;
    x2 = round((double)((int)item->chromEnd-winStart)*scale) + xOff;
    w = x2-x1;
    color = shadesOfGray[grayInRange(item->gcPpt, 340, 617)];
    if (w < 1)
	w = 1;
    mgDrawBox(mg, x1, y, w, heightPer, color);
    printf("<AREA SHAPE=RECT COORDS=\"%d,%d,%d,%d\" ", x1, y, x1+w, y+heightPer);
    printf("HREF=\"../cgi-bin/hgc?g=%s&i=%s&c=%s&l=%d&r=%d&o=%d\" ", 
	tg->mapName, item->name, chromName, winStart, winEnd, item->chromStart);
    printf("ALT= \"%s\" TITLE=\"%s\">\n", item->name, item->name); 
    if (isFull)
	y += lineHeight;
    }
}



struct trackGroup *isochoresTg()
/* Make track group for isochores. */
{
struct trackGroup *tg;

AllocVar(tg);
tg->mapName = "hgIsochore";
tg->visibility = tvHide;
tg->longLabel = "GC-rich (dark) and AT-rich (light) isochores";
tg->shortLabel = "isochores";
tg->loadItems = isochoreLoad;
tg->freeItems = isochoreFree;
tg->drawItems = isochoreDraw;
tg->ignoresColor = TRUE;
tg->itemName = isochoreName;
tg->mapItemName = bedName;
tg->totalHeight = tgFixedTotalHeight;
tg->itemHeight = tgFixedItemHeight;
return tg;
}

char *simpleRepeatName(struct trackGroup *tg, void *item)
/* Return name of simpleRepeats track item. */
{
struct simpleRepeat *rep = item;
static char buf[17];
int len;

if (rep->sequence == NULL)
    return rep->name;
len = strlen(rep->sequence);
if (len <= 16)
    return rep->sequence;
else
    {
    memcpy(buf, rep->sequence, 13);
    memcpy(buf+13, "...", 3);
    buf[16] = 0;
    return buf;
    }
}

void loadSimpleRepeats(struct trackGroup *tg)
/* Load up simpleRepeats from database table to trackGroup items. */
{
bedLoad(tg, "simpleRepeat", (ItemLoader)simpleRepeatLoad);
}

void freeSimpleRepeats(struct trackGroup *tg)
/* Free up isochore items. */
{
simpleRepeatFreeList((struct simpleRepeat**)&tg->items);
}

struct trackGroup *simpleRepeatTg()
/* Make track group for simple repeats. */
{
struct trackGroup *tg;

AllocVar(tg);
tg->mapName = "hgSimpleRepeat";
tg->visibility = tvHide;
tg->longLabel = "Simple Tandem Repeats";
tg->shortLabel = "Simple Repeats";
tg->loadItems = loadSimpleRepeats;
tg->freeItems = freeSimpleRepeats;
tg->drawItems = bedDrawSimple;
tg->itemName = simpleRepeatName;
tg->mapItemName = bedName;
tg->totalHeight = tgFixedTotalHeight;
tg->itemHeight = tgFixedItemHeight;
return tg;
}

void loadCpgIsland(struct trackGroup *tg)
/* Load up simpleRepeats from database table to trackGroup items. */
{
bedLoad(tg, "cpgIsland", (ItemLoader)cpgIslandLoad);
}

void freeCpgIsland(struct trackGroup *tg)
/* Free up isochore items. */
{
cpgIslandFreeList((struct cpgIsland**)&tg->items);
}

struct trackGroup *cpgIslandTg()
/* Make track group for simple repeats. */
{
struct trackGroup *tg;

AllocVar(tg);
tg->mapName = "hgCpgIsland";
tg->visibility = tvHide;
tg->longLabel = "CpG Islands with length > 50bp and GC > 50%";
tg->shortLabel = "CpG Islands";
tg->loadItems = loadCpgIsland;
tg->freeItems = freeCpgIsland;
tg->drawItems = bedDrawSimple;
tg->itemName = bedName;
tg->mapItemName = bedName;
tg->totalHeight = tgFixedTotalHeight;
tg->itemHeight = tgFixedItemHeight;
tg->color.r = 0;
tg->color.g = 100;
tg->color.b = 0;
return tg;
}

char *cytoBandName(struct trackGroup *tg, void *item)
/* Return name of cytoBand track item. */
{
struct cytoBand *band = item;
static char buf[32];
int len;

sprintf(buf, "%s%s", skipChr(band->chrom), band->name);
return buf;
}

char *abbreviatedBandName(struct trackGroup *tg, struct cytoBand *band, MgFont *font, int width)
/* Return a string abbreviated enough to fit into space. */
{
int textWidth;
static char string[32];

/* If have enough space, return original chromosome/band. */
sprintf(string, "%s%s", skipChr(band->chrom), band->name);
textWidth = mgFontStringWidth(font, string);
if (textWidth <= width)
    return string;

/* Try leaving off chromosome  */
sprintf(string, "%s", band->name);
textWidth = mgFontStringWidth(font, string);
if (textWidth <= width)
    return string;

/* Try leaving off initial letter */
sprintf(string, "%s", band->name+1);
textWidth = mgFontStringWidth(font, string);
if (textWidth <= width)
    return string;

return NULL;
}

void cytoBandColor(struct cytoBand *band, Color cenColor, Color *retBoxColor, Color *retTextColor)
/* Figure out color of band. */
{
char *stain = band->gieStain;
if (startsWith("gneg", stain))
    {
    *retBoxColor = shadesOfGray[1];
    *retTextColor = MG_BLACK;
    }
else if (startsWith("gpos", stain))
    {
    int percentage = 100;
    stain += 4;	
    if (isdigit(stain[0]))
        percentage = atoi(stain);
    *retBoxColor = shadesOfGray[grayInRange(percentage, -30, 100)];
    *retTextColor = MG_WHITE;
    }
else if (startsWith("gvar", stain))
    {
    *retBoxColor = shadesOfGray[2];
    *retTextColor = MG_WHITE;
    }
else 
    {
    *retBoxColor = cenColor;
    *retTextColor = MG_WHITE;
    }
}

static void cytoBandDraw(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
/* Draw cytoBand items. */
{
int baseWidth = seqEnd - seqStart;
struct cytoBand *band;
int y = yOff;
int heightPer = tg->heightPer;
int lineHeight = tg->lineHeight;
int x1,x2,w;
int midLineOff = heightPer/2;
boolean isFull = (vis == tvFull);
Color col, textCol;
int ix = 0;
char *s;
double scale = width/(double)baseWidth;
for (band = tg->items; band != NULL; band = band->next)
    {
    x1 = round((double)((int)band->chromStart-winStart)*scale) + xOff;
    x2 = round((double)((int)band->chromEnd-winStart)*scale) + xOff;
    /* Clip here so that text will tend to be more visible... */
    if (x1 < xOff)
	x1 = xOff;
    if (x2 > xOff + width)
	x2 = xOff + width;
    w = x2-x1;
    if (w < 1)
	w = 1;
    cytoBandColor(band, tg->ixAltColor, &col, &textCol);
    mgDrawBox(mg, x1, y, w, heightPer, col);
    s = abbreviatedBandName(tg, band, tl.font, w);
    if (s != NULL)
	mgTextCentered(mg, x1, y, w, heightPer, textCol, tl.font, s);
    mapBoxJumpTo(x1, y, w, heightPer, band->chrom, band->chromStart, band->chromEnd, cytoBandName(tg, band));
    if (isFull)
	y += lineHeight;
    ++ix;
    }
}


void loadCytoBands(struct trackGroup *tg)
/* Load up simpleRepeats from database table to trackGroup items. */
{
bedLoad(tg, "cytoBand", (ItemLoader)cytoBandLoad);
}

void freeCytoBands(struct trackGroup *tg)
/* Free up isochore items. */
{
cytoBandFreeList((struct cytoBand**)&tg->items);
}

struct trackGroup *cytoBandTg()
/* Make track group for simple repeats. */
{
struct trackGroup *tg;

AllocVar(tg);
tg->mapName = "hgCytoBands";
tg->visibility = tvDense;
tg->longLabel = "Chromosome Band (click on band to zoom to that band)";
tg->shortLabel = "Chromosome Band";
tg->loadItems = loadCytoBands;
tg->freeItems = freeCytoBands;
tg->drawItems = cytoBandDraw;
tg->itemName = cytoBandName;
tg->mapItemName = bedName;
tg->totalHeight = tgFixedTotalHeight;
tg->itemHeight = tgFixedItemHeight;
tg->altColor.r = 150;
tg->altColor.g = 50;
tg->altColor.b = 50;
return tg;
}

void loadGcPercent(struct trackGroup *tg)
/* Load up simpleRepeats from database table to trackGroup items. */
{
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr = NULL;
char **row;
struct gcPercent *itemList = NULL, *item;
char query[256];

sprintf(query, "select * from %s where chrom = '%s' and chromStart<%u and chromEnd>%u",
    "gcPercent", chromName, winEnd, winStart);

/* Get the frags and load into tg->items. */
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    item = gcPercentLoad(row);
    if (item->gcPpt != 0)
	{
	slAddHead(&itemList, item);
	}
    else
        gcPercentFree(&item);
    }
slReverse(&itemList);
sqlFreeResult(&sr);
tg->items = itemList;
hFreeConn(&conn);
}

void freeGcPercent(struct trackGroup *tg)
/* Free up isochore items. */
{
gcPercentFreeList((struct gcPercent**)&tg->items);
}

char *gcPercentName(struct trackGroup *tg, void *item)
/* Return name of gcPercent track item. */
{
struct gcPercent *gc = item;
static char buf[32];

sprintf(buf, "%3.1f%% GC", 0.1*gc->gcPpt);
return buf;
}

static int gcPercentMin = 320;
static int gcPercentMax = 600;

Color gcPercentColor(struct trackGroup *tg, void *item)
/* Return name of gcPercent track item. */
{
struct gcPercent *gc = item;
int ppt = gc->gcPpt;
int grayLevel;

grayLevel = grayInRange(ppt, gcPercentMin, gcPercentMax);
return shadesOfGray[grayLevel];
}

static void gcPercentDenseDraw(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
/* Draw gcPercent items. */
{
int baseWidth = seqEnd - seqStart;
UBYTE *useCounts;
UBYTE *aveCounts;
int i;
int lineHeight = mgFontLineHeight(font);
struct gcPercent *gc;
int s, e, w;
int log2 = digitsBaseTwo(baseWidth);
int shiftFactor = log2 - 17;
int sampleWidth;

if (shiftFactor < 0)
    shiftFactor = 0;
sampleWidth = (baseWidth>>shiftFactor);
AllocArray(useCounts, sampleWidth);
AllocArray(aveCounts, width);
memset(useCounts, 0, sampleWidth * sizeof(useCounts[0]));
for (gc = tg->items; gc != NULL; gc = gc->next)
    {
    s = ((gc->chromStart - seqStart)>>shiftFactor);
    e = ((gc->chromStart - seqStart)>>shiftFactor);
    if (s < 0) s = 0;
    if (e > sampleWidth) e = sampleWidth;
    w = e - s;
    if (w > 0)
	memset(useCounts+s, grayInRange(gc->gcPpt, gcPercentMin, gcPercentMax), w);
    }
resampleBytes(useCounts, sampleWidth, aveCounts, width);
grayThreshold(aveCounts, width);
for (i=0; i<lineHeight; ++i)
    mgPutSegZeroClear(mg, xOff, yOff+i, width, aveCounts);
freeMem(useCounts);
freeMem(aveCounts);
}

static void gcPercentDraw(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
/* Draw gcPercent items. */
{
if (vis == tvDense)
   gcPercentDenseDraw(tg, seqStart, seqEnd, mg, xOff, yOff, width, font, color, vis);
else
   bedDrawSimple(tg, seqStart, seqEnd, mg, xOff, yOff, width, font, color, vis);
}


struct trackGroup *gcPercentTg()
/* Make track group for simple repeats. */
{
struct trackGroup *tg;

AllocVar(tg);
tg->mapName = "hgGcPercent";
tg->visibility = tvHide;
tg->longLabel = "Percentage GC in 20,000 Base Windows";
tg->shortLabel = "GC Percent";
tg->loadItems = loadGcPercent;
tg->freeItems = freeGcPercent;
tg->drawItems = bedDrawSimple;
tg->itemName = gcPercentName;
tg->mapItemName = bedName;
tg->totalHeight = tgFixedTotalHeight;
tg->itemHeight = tgFixedItemHeight;
tg->ignoresColor = TRUE;
tg->itemColor = gcPercentColor;
return tg;
}

void loadGenomicDups(struct trackGroup *tg)
/* Load up simpleRepeats from database table to trackGroup items. */
{
bedLoad(tg, "genomicDups", (ItemLoader)genomicDupsLoad);
}

void freeGenomicDups(struct trackGroup *tg)
/* Free up isochore items. */
{
genomicDupsFreeList((struct genomicDups**)&tg->items);
}

Color genomicDupsColor(struct trackGroup *tg, void *item)
/* Return name of gcPercent track item. */
{
struct genomicDups *dup = item;
int ppt = dup->score;
int grayLevel;

if (ppt > 990)
    return tg->ixColor;
else if (ppt > 980)
    return tg->ixAltColor;
grayLevel = grayInRange(ppt, 900, 1000);
return shadesOfGray[grayLevel];
}

char *genomicDupsName(struct trackGroup *tg, void *item)
/* Return full genie name. */
{
struct genomicDups *gd = item;
char *full = gd->name;
static char abbrev[64];
int len;
char *e;

strcpy(abbrev, skipChr(full));
abbr(abbrev, "om");
return abbrev;
}



struct trackGroup *genomicDupsTg()
/* Make track group for simple repeats. */
{
struct trackGroup *tg;

AllocVar(tg);
tg->mapName = "hgGenomicDups";
tg->visibility = tvDense;
tg->longLabel = "Duplications of >1000 bases of non-RepeatMasked Sequence";
tg->shortLabel = "duplications";
tg->loadItems = loadGenomicDups;
tg->freeItems = freeGenomicDups;
tg->drawItems = bedDrawSimple;
tg->itemName = genomicDupsName;
tg->mapItemName = bedName;
tg->totalHeight = tgFixedTotalHeight;
tg->itemHeight = tgFixedItemHeight;
tg->color.r = 250;
tg->color.g = 130;
tg->color.b = 0;
tg->altColor.r = 230;
tg->altColor.g = 200;
tg->altColor.b = 0;
tg->itemColor = genomicDupsColor;
return tg;
}

void loadGenethon(struct trackGroup *tg)
/* Load up simpleRepeats from database table to trackGroup items. */
{
bedLoad(tg, "mapGenethon", (ItemLoader)mapStsLoad);
}

void freeGenethon(struct trackGroup *tg)
/* Free up isochore items. */
{
mapStsFreeList((struct mapSts**)&tg->items);
}

struct trackGroup *genethonTg()
/* Make track group for simple repeats. */
{
struct trackGroup *tg;

AllocVar(tg);
tg->mapName = "hgGenethon";
tg->visibility = tvDense;
tg->longLabel = "Genetic Markers From Genethon";
tg->shortLabel = "Genethon";
tg->loadItems = loadGenethon;
tg->freeItems = freeGenethon;
tg->drawItems = bedDrawSimple;
tg->itemName = bedName;
tg->mapItemName = bedName;
tg->totalHeight = tgFixedTotalHeight;
tg->itemHeight = tgFixedItemHeight;
return tg;
}

void loadEst3(struct trackGroup *tg)
/* Load up simpleRepeats from database table to trackGroup items. */
{
bedLoad(tg, "est3", (ItemLoader)est3Load);
}

void freeEst3Items(struct trackGroup *tg)
/* Free up isochore items. */
{
est3FreeList((struct est3**)&tg->items);
}

char *est3Name(struct trackGroup *tg, void *item)
/* Return what to display on left column of open track. */
{
struct est3 *e3 = item;
static char name[64];

sprintf(name, "%d %s 3' ESTs", e3->estCount, e3->strand);
return name;
}

char *est3MapName(struct trackGroup *tg, void *item)
/* Return what to write in client-side image map. */
{
return "est3";
}


struct trackGroup *est3Tg()
/* Make track group for simple repeats. */
{
struct trackGroup *tg;

AllocVar(tg);
tg->mapName = "hgEst3";
tg->visibility = tvDense;
tg->longLabel = "EST 3' Ends, Filtered and Clustered";
tg->shortLabel = "EST 3' Ends";
tg->loadItems = loadEst3;
tg->freeItems = freeEst3Items;
tg->drawItems = bedDrawSimple;
tg->itemName = est3Name;
tg->mapItemName = est3MapName;
tg->totalHeight = tgFixedTotalHeight;
tg->itemHeight = tgFixedItemHeight;
return tg;
}

#ifdef EXAMPLE
void loadXyz(struct trackGroup *tg)
/* Load up simpleRepeats from database table to trackGroup items. */
{
bedLoad(tg, "xyz", (ItemLoader)xyzLoad);
}

void freeXyz(struct trackGroup *tg)
/* Free up isochore items. */
{
xyzFreeList((struct xyz**)&tg->items);
}

struct trackGroup *xyzTg()
/* Make track group for simple repeats. */
{
struct trackGroup *tg;

AllocVar(tg);
tg->mapName = "hgXyz";
tg->visibility = tvDense;
tg->longLabel = "xyz";
tg->shortLabel = "xyz";
tg->loadItems = loadXyz;
tg->freeItems = freeXyz;
tg->drawItems = bedDrawSimple;
tg->itemName = bedName;
tg->mapItemName = bedName;
tg->totalHeight = tgFixedTotalHeight;
tg->itemHeight = tgFixedItemHeight;
return tg;
}
#endif /* EXAMPLE */

struct wabaChromHit
/* Records where waba alignment hits chromosome. */
    {
    struct wabaChromHit *next;	/* Next in list. */
    char *query;	  	/* Query name. */
    int chromStart, chromEnd;   /* Chromosome position. */
    char strand;                /* + or - for strand. */
    int milliScore;             /* Parts per thousand */
    char *squeezedSym;          /* HMM Symbols */
    };

struct wabaChromHit *wchLoad(char *row[])
/* Create a wabaChromHit from database row. 
 * Since squeezedSym autoSql can't generate this,
 * alas. */
{
int size;
char *sym;
struct wabaChromHit *wch;

AllocVar(wch);
wch->query = cloneString(row[0]);
wch->chromStart = sqlUnsigned(row[1]);
wch->chromEnd = sqlUnsigned(row[2]);
wch->strand = row[3][0];
wch->milliScore = sqlUnsigned(row[4]);
size = wch->chromEnd - wch->chromStart;
wch->squeezedSym = sym = needLargeMem(size+1);
memcpy(sym, row[5], size);
sym[size] = 0;
return wch;
}

void wchFree(struct wabaChromHit **pWch)
/* Free a singlc wabaChromHit. */
{
struct wabaChromHit *wch = *pWch;
if (wch != NULL)
    {
    freeMem(wch->squeezedSym);
    freeMem(wch->query);
    freez(pWch);
    }
}

void wchFreeList(struct wabaChromHit **pList)
/* Free list of wabaChromHits. */
{
struct wabaChromHit *el, *next;

for (el = *pList; el != NULL; el = next)
    {
    next = el->next;
    wchFree(&el);
    }
*pList = NULL;
}

void wabaLoad(struct trackGroup *tg)
/* Load up waba items intersecting window. */
{
char table[64];
char query[256];
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr = NULL;
char **row;
struct agpFrag *fragList = NULL, *frag;
struct agpGap *gapList = NULL, *gap;
struct wabaChromHit *wch, *wchList = NULL;

/* Get the frags and load into tg->items. */
sprintf(table, "%s%s", chromName, tg->customPt);
sprintf(query, "select * from %s where chromStart<%u and chromEnd>%u",
    table, winEnd, winStart);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    wch = wchLoad(row);
    slAddHead(&wchList, wch);
    }
slReverse(&wchList);
tg->items = wchList;
sqlFreeResult(&sr);
hFreeConn(&conn);
}

void wabaFree(struct trackGroup *tg)
/* Free up wabaTrackGroup items. */
{
wchFreeList((struct wabaChromHit**)&tg->items);
}


void makeSymColors(struct trackGroup *tg, enum trackVisibility vis,
	Color symColor[128])
/* Fill in array with color for each symbol value. */
{
memset(symColor, MG_WHITE, 128);
symColor['1'] = symColor['2'] = symColor['3'] = symColor['H'] 
   = tg->ixColor;
symColor['L'] = tg->ixAltColor;
}

int countSameColor(char *sym, int symCount, Color symColor[])
/* Count how many symbols are the same color as the current one. */
{
Color color = symColor[sym[0]];
int ix;
for (ix = 1; ix < symCount; ++ix)
    {
    if (symColor[sym[ix]] != color)
        break;
    }
return ix;
}

void wabaItemMap(struct trackGroup *tg, void *item, int x, int y, int width, int height)
/* Print image map line on one waba item. */
{
struct wabaChromHit *wch = item;
char *id = tg->mapName;
char *shortName = cgiEncode(tg->mapItemName(tg, item));
char *statusLine = cgiEncode(tg->itemName(tg, item));

printf("<AREA SHAPE=RECT COORDS=\"%d,%d,%d,%d\" ", x, y, x+width, y+height);
printf("HREF=\"../cgi-bin/hgc?g=%s&i=%s&c=%s&l=%d&r=%d&o=%d\" ", 
    id, shortName, chromName, winStart, winEnd, wch->chromStart);
printf("ALT= \"%s\" TITLE=\"%s\">\n", statusLine, statusLine); 
freeMem(shortName);
freeMem(statusLine);
}


static void wabaDraw(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
/* Draw waba alignment items. */
{
Color symColor[128];
int baseWidth = seqEnd - seqStart;
struct wabaChromHit *wch;
int y = yOff;
int heightPer = tg->heightPer;
int lineHeight = tg->lineHeight;
int x1,x2,w;
boolean isFull = (vis == tvFull);
int ix = 0;


makeSymColors(tg, vis, symColor);
for (wch = tg->items; wch != NULL; wch = wch->next)
    {
    int chromStart = wch->chromStart;
    int symCount = wch->chromEnd - chromStart;
    int typeStart, typeWidth, typeEnd;
    char *sym = wch->squeezedSym;
    for (typeStart = 0; typeStart < symCount; typeStart = typeEnd)
	{
	typeWidth = countSameColor(sym+typeStart, symCount-typeStart, symColor);
	typeEnd = typeStart + typeWidth;
	color = symColor[sym[typeStart]];
	if (color != MG_WHITE)
	    {
	    x1 = roundingScale(typeStart+chromStart-winStart, width, baseWidth)+xOff;
	    x2 = roundingScale(typeEnd+chromStart-winStart, width, baseWidth)+xOff;
	    w = x2-x1;
	    if (w < 1)
		w = 1;
	    mgDrawBox(mg, x1, y, w, heightPer, color);
	    }
	}
    if (isFull)
	y += lineHeight;
    }
}

char *wabaName(struct trackGroup *tg, void *item)
/* Return name of waba track item. */
{
struct wabaChromHit *wch = item;
return wch->query;
}

struct trackGroup *wabaTrackGroup()
/* Return track with fields shared by waba-based 
 * alignment tracks filled in. */
{
struct trackGroup *tg;
AllocVar(tg);
tg->loadItems = wabaLoad;
tg->freeItems = wabaFree;
tg->drawItems = wabaDraw;
tg->itemName = wabaName;
tg->mapItemName = wabaName;
tg->totalHeight = tgFixedTotalHeight;
tg->itemHeight = tgFixedItemHeight;
tg->itemMap = wabaItemMap;
return tg;
}

struct trackGroup *tetTg()
/* Make track group for Tetraodon alignments. */
{
struct trackGroup *tg;

tg = wabaTrackGroup();
tg->mapName = "hgTet";
tg->visibility = tvDense;
tg->longLabel = "Tetraodon Nigroviridis Homologies";
tg->shortLabel = "Tetraodon";
tg->color.r = 50;
tg->color.g = 100;
tg->color.b = 200;
tg->altColor.r = 85;
tg->altColor.g = 170;
tg->altColor.b = 225;
tg->customPt = "_tet_waba";
return tg;
}

void contigLoad(struct trackGroup *tg)
/* Load up contigs from database table to trackGroup items. */
{
char query[256];
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr = NULL;
char **row;
struct ctgPos *ctgList = NULL, *ctg;

/* Get the contigs and load into tg->items. */
sprintf(query, "select * from ctgPos where chrom = '%s' and chromStart<%u and chromEnd>%u",
    chromName, winEnd, winStart);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    ctg = ctgPosLoad(row);
    slAddHead(&ctgList, ctg);
    }
slReverse(&ctgList);
sqlFreeResult(&sr);
hFreeConn(&conn);
tg->items = ctgList;
}

char *abbreviateContig(char *string, MgFont *font, int width)
/* Return a string abbreviated enough to fit into space. */
{
int textWidth;

/* If have enough space, return original unabbreviated string. */
textWidth = mgFontStringWidth(font, string);
if (textWidth <= width)
    return string;

/* Try skipping over 'ctg' */
string += 3;
textWidth = mgFontStringWidth(font, string);
if (textWidth <= width)
    return string;
return NULL;
}

static void contigDraw(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
/* Draw contig items. */
{
int baseWidth = seqEnd - seqStart;
struct ctgPos *ctg;
int y = yOff;
int heightPer = tg->heightPer;
int lineHeight = tg->lineHeight;
int x1,x2,w;
int midLineOff = heightPer/2;
boolean isFull = (vis == tvFull);
Color col;
int ix = 0;
char *s;
double scale = width/(double)baseWidth;
for (ctg = tg->items; ctg != NULL; ctg = ctg->next)
    {
    x1 = round((double)((int)ctg->chromStart-winStart)*scale) + xOff;
    x2 = round((double)((int)ctg->chromEnd-winStart)*scale) + xOff;
    /* Clip here so that text will tend to be more visible... */
    if (x1 < xOff)
	x1 = xOff;
    if (x2 > xOff + width)
	x2 = xOff + width;
    w = x2-x1;
    if (w < 1)
	w = 1;
    mgDrawBox(mg, x1, y, w, heightPer, color);
    s = abbreviateContig(ctg->contig, tl.font, w);
    if (s != NULL)
	mgTextCentered(mg, x1, y, w, heightPer, MG_WHITE, tl.font, s);
    if (isFull)
	y += lineHeight;
    else 
	{
	mapBoxHc(x1,y,w,heightPer, tg->mapName, 
	    tg->mapItemName(tg, ctg), 
	    tg->itemName(tg, ctg));
	}
    ++ix;
    }
}


void contigFree(struct trackGroup *tg)
/* Free up contigTrackGroup items. */
{
ctgPosFreeList((struct ctgPos**)&tg->items);
}


char *contigName(struct trackGroup *tg, void *item)
/* Return name of contig track item. */
{
struct ctgPos *ctg = item;
return ctg->contig;
}

struct trackGroup *contigTg()
/* Make track group for contig */
{
struct trackGroup *tg;

AllocVar(tg);
tg->mapName = "hgContig";
tg->visibility = tvHide;
tg->longLabel = "Fingerprint Map Contigs";
tg->shortLabel = "FPC Contigs";
tg->loadItems = contigLoad;
tg->freeItems = contigFree;
tg->drawItems = contigDraw;
tg->color.r = 150;
tg->color.g = 0;
tg->color.b = 0;
tg->itemName = contigName;
tg->mapItemName = contigName;
tg->totalHeight = tgFixedTotalHeight;
tg->itemHeight = tgFixedItemHeight;
return tg;
}

struct cloneFrag
/* A fragment of a clone. */
    {
    struct cloneFrag *next;	/* Next in list. */
    char *name;			/* Name of fragment.  Not allocated here. */
    };

struct cloneFragPos
/* An alignment involving a clone frag. */
    {
    struct cloneFragPos *next;	/* Next in list. */
    struct cloneFrag *frag;     /* Fragment info. */
    struct psl *psl;            /* Alignment info. Memory owned here. Possibly NULL. */
    struct gl *gl;              /* Golden path position info. */
    int start, end;             /* Start end in chromosome. */
    };

struct cloneInfo
/* Info about a clone and where it aligns. */
    {
    struct cloneInfo *next; 	/* Next in list */
    char *name;                 /* Name of clone. (Not allocated here) */
    short phase;			/* Htg phase - 1 2 or 3. */
    char stage;                 /* Stage - (P)redraft, (D)raft, (F)inished. */
    struct cloneFrag *fragList; /* List of fragments. */
    int fragCount;              /* Count of fragments. */
    struct cloneFragPos *cfaList;   /* List of alignments. */
    struct spaceSaver *ss;      /* How this is layed out. */
    int cloneStart, cloneEnd;       /* Min/Max position of alignments. */
    };

int cmpCloneInfo(const void *va, const void *vb)
/* Compare two cloneInfos by cloneStart. */
{
const struct cloneInfo *a = *((struct cloneInfo **)va);
const struct cloneInfo *b = *((struct cloneInfo **)vb);
return a->cloneStart - b->cloneStart;
}


void cloneInfoFree(struct cloneInfo **pCi)
/* free up a clone info. */
{
struct cloneInfo *ci;
if ((ci = *pCi) != NULL)
    {
    struct cloneFragPos *cfa;
    for (cfa = ci->cfaList; cfa != NULL; cfa = cfa->next)
	{
	pslFree(&cfa->psl);
	glFree(&cfa->gl);
	}
    slFreeList(&ci->cfaList);
    slFreeList(&ci->fragList);
    freez(pCi);
    }
}

void cloneInfoFreeList(struct cloneInfo **pList)
/* Free up a list of cloneInfo's. */
{
struct cloneInfo *el,*next;
for (el = *pList; el != NULL; el = next)
    {
    next = el->next;
    cloneInfoFree(&el);
    }
*pList = NULL;
}


struct cloneFrag *findCloneFrag(struct cloneInfo *ci, char *fragName)
/* Search for named frag and return it, or NULL if not found. */
{
struct cloneFrag *frag;
for (frag = ci->fragList; frag != NULL; frag = frag->next)
    {
    if (sameString(frag->name, fragName))
	return frag;
    }
return NULL;
}

void layoutCloneAli(struct cloneInfo *ci)
/* Figure out space saver layout for alignments in clone. */
{
struct spaceSaver *ss;
struct cloneFragPos *cfa;
int start = 0x3fffffff;
int end = 0;

ss = ci->ss = spaceSaverNew(winStart, winEnd);
for (cfa = ci->cfaList; cfa != NULL; cfa = cfa->next)
    {
    spaceSaverAdd(ss, cfa->start, cfa->end, cfa);
    }
spaceSaverFinish(ss);
}

char *cloneName(struct trackGroup *tg, void *item)
/* Return name of gold track item. */
{
struct cloneInfo *ci = item;
return ci->name;
}

int cloneFragMaxWin = 1500000;

int oneOrRowCount(struct cloneInfo *ci)
/* Return row count, but at least one. */
{
int rowCount = ci->ss->rowCount;
if (rowCount < 1) rowCount = 1;
return rowCount;
}

static int cloneItemHeight(struct trackGroup *tg, void *item)
/* Return item height for fixed height track. */
{
struct cloneInfo *ci = item;
int height1 = mgFontLineHeight(tl.font)+1;
if (winBaseCount <= cloneFragMaxWin)
    return height1*oneOrRowCount(ci)+2;
else
    return height1;
}

static int cloneTotalHeight(struct trackGroup *tg, enum trackVisibility vis)
/* Height of a clone track. */
{
switch (vis)
    {
    case tvFull:
	{
	int total = 0;
	struct cloneInfo *ci;
	for (ci = tg->items; ci != NULL; ci = ci->next)
	    {
	    total += tg->itemHeight(tg, ci);
	    }
	tg->height = total+2;
	break;
	}
    case tvDense:
	tg->lineHeight = mgFontLineHeight(tl.font)+1;
	tg->heightPer = tg->lineHeight - 1;
	tg->height = tg->lineHeight;
	break;
    }
return tg->height;
}

static void drawOneClone(struct cloneInfo *ci, int seqStart, int seqEnd,
    struct memGfx *mg, int xOff, int yOff, int width,
    MgFont *font, int lineHeight, Color color, boolean stagger, 
    boolean hiliteDupes)
/* Draw a single clone item - using space saver layout on fragments. */
{
struct cloneFragPos *cfa;
struct psl *psl;
int y;
int heightPer = lineHeight-1;
struct spaceSaver *ss = ci->ss;
int baseWidth = seqEnd - seqStart;
struct spaceNode *sn;
int x1, x2, w;
char *s;
int textWidth;
char fullPos[256];
struct hash *dupeHash = NULL;
Color col;
struct hashEl *hel;

if (hiliteDupes)
    {
    struct hash *uniqHash = newHash(7);
    dupeHash = newHash(6);
    for (sn = ss->nodeList; sn != NULL; sn = sn->next)
	{
	cfa = sn->val;
	if (cfa->end - cfa->start > 1000)
	    {
	    s = strchr(cfa->frag->name, '_');
	    if (s != NULL)
		{
		s += 1;
		if (hashLookup(uniqHash, s) == NULL)
		    {
		    hashAdd(uniqHash, s, NULL);
		    }
		else	/* Got a dupe! */
		    {
		    if ((hel = hashLookup(dupeHash, s)) == NULL)
			hashAdd(dupeHash, s, sn);
		    }
		}
	    }
	}
    freeHash(&uniqHash);
    }

for (sn = ss->nodeList; sn != NULL; sn = sn->next)
    {
    if (stagger)
	y = yOff + sn->row*lineHeight;
    else
	y = yOff;
    cfa = sn->val;
    x1 = roundingScale(cfa->start-winStart, width, baseWidth)+xOff;
    x2 = roundingScale(cfa->end-winStart, width, baseWidth)+xOff;

    /* Clip here so that text will tend to be more visible... */
    if (x1 < xOff)
	x1 = xOff;
    if (x2 > xOff + width)
	x2 = xOff + width;
    w = x2-x1;
    if (w < 1)
	w = 1;
    s = strchr(cfa->frag->name, '_');
    if (s == NULL)
	s = "";
    else
	s += 1;
    col = color;
    if (hiliteDupes)
	{
	if ((hel = hashLookup(dupeHash, s)) != NULL)
	    {
	    if (hel->val == sn)
		col = MG_RED;
	    else
		col = MG_BLUE;
	    }
	}
    mgDrawBox(mg, x1, y, w, heightPer, col);
    textWidth = mgFontStringWidth(font, s);
    if (textWidth <= w)
	mgTextCentered(mg, x1, y, w, heightPer, MG_WHITE, font, s);
    if (baseWidth <= 2000000)
	{
	psl = cfa->psl;
	if (psl != NULL)
	    {
	    sprintf(fullPos, "%s %d to %d of %d, strand %s, hits %d to %d", 
	    	psl->qName, psl->qStart, 
		psl->qEnd, psl->qSize, psl->strand,
		psl->tStart, psl->tEnd);
	    mapBoxHc(x1,y,w,heightPer, "hgClone", cfa->frag->name, fullPos);
	    }
	else
	    mapBoxHc(x1,y,w,heightPer, "hgClone", cfa->frag->name, cfa->frag->name);
	}
    }
freeHash(&dupeHash);
}

/* These tables are for combining sequence scores. 
 *    0 = no coverage
 *    1 = predraft
 *    2 = draft
 *    3 = deep draft
 *    4 = finished  */
static UBYTE predraftInc[5] = {1, 1, 2, 3, 4};  
static UBYTE draftInc[5] = {2, 2, 3, 3, 4};
static UBYTE finishedInc[5] = {4, 4, 4, 4, 4};

void incStage(UBYTE *b, int width, char stage)
/* Increment b from 0 to width-1 according to stage. */
{
UBYTE *inc;
int i;

if (stage == 'P')
    inc = predraftInc;
else if (stage == 'D')
    inc = draftInc;
else if (stage == 'F')
    inc = finishedInc;
else
    errAbort("Unknown stage %c", stage);
for (i=0; i<width; ++i)
   b[i] = inc[b[i]];
}

static void cloneDenseDraw(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
/* Draw dense clone items. */
{
int baseWidth = seqEnd - seqStart;
UBYTE *useCounts;
UBYTE *aveCounts;
int i;
int lineHeight = mgFontLineHeight(font);
struct cloneInfo *ci;
struct cloneFragPos *cfa;   /* List of alignments. */
int s, e, w;
int log2 = digitsBaseTwo(baseWidth);
int shiftFactor = log2 - 17;
int sampleWidth;

if (shiftFactor < 0)
    shiftFactor = 0;
sampleWidth = (baseWidth>>shiftFactor);
AllocArray(useCounts, sampleWidth);
AllocArray(aveCounts, width);
memset(useCounts, 0, sampleWidth * sizeof(useCounts[0]));
for (ci = tg->items; ci != NULL; ci = ci->next)
    {
    char stage = ci->stage;
    for (cfa = ci->cfaList; cfa != NULL; cfa = cfa->next)
	{
	s = ((cfa->start - seqStart)>>shiftFactor);
	e = ((cfa->end - seqStart)>>shiftFactor);
	if (s < 0) s = 0;
	if (e > sampleWidth) e = sampleWidth;
	w = e - s;
	if (w > 0)
	    incStage(useCounts+s, w, stage);
	}
    }
resampleBytes(useCounts, sampleWidth, aveCounts, width);
grayThreshold(aveCounts, width);
for (i=0; i<lineHeight; ++i)
    mgPutSegZeroClear(mg, xOff, yOff+i, width, aveCounts);
freeMem(useCounts);
freeMem(aveCounts);
}

static void cloneFullDraw(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
/* Draw full  clone items. */
{
int y = yOff;
int lineHeight = mgFontLineHeight(font)+1;
struct cloneInfo *ci;
Color light = tg->ixAltColor;
int oneHeight;
int x1, x2, w;
int baseWidth = seqEnd - seqStart;
int tooBig = (winBaseCount > cloneFragMaxWin);


for (ci = tg->items; ci != NULL; ci = ci->next)
    {
    if (!tooBig)
	oneHeight = oneOrRowCount(ci)*lineHeight+2;
    else
	oneHeight = lineHeight;
    x1 = roundingScale(ci->cloneStart-winStart, width, baseWidth)+xOff;
    x2 = roundingScale(ci->cloneEnd-winStart, width, baseWidth)+xOff;
    w = x2-x1;
    mgDrawBox(mg, x1, y, w, oneHeight-1, light);
    if (!tooBig)
	drawOneClone(ci, seqStart, seqEnd, mg, xOff, y+1, width, font, lineHeight, 
		color, TRUE, tg->subType);
    else
	drawOneClone(ci, seqStart, seqEnd, mg, xOff, y, width, font, oneHeight-1, 
		color, FALSE, tg->subType);
    y += oneHeight;
    }
}

static void cloneDraw(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
/* Draw clone items. */
{
if (vis == tvFull)
    cloneFullDraw(tg, seqStart, seqEnd, mg, xOff, yOff, width, font, color, vis);
else
    cloneDenseDraw(tg, seqStart, seqEnd, mg, xOff, yOff, width, font, color, vis);
}


struct hash *realiCloneHash;
struct cloneInfo *realiCloneList;

void loadRealiClonesInWindow()
/* Load up realiCloneHash and realiCloneList with the clones in the window. */
{
if (realiCloneList == NULL)
    {
    char query[256];
    struct sqlConnection *conn = hAllocConn();
    struct sqlResult *sr = NULL;
    char **row;
    struct cloneInfo *ci;
    struct psl *psl;
    char *fragName;
    struct cloneFrag *cf;
    char cloneName[128];
    struct hashEl *hel;
    struct cloneFragPos *cfa;
    char *s;
    struct clonePos cp;

    /* Load in clone extents from database. */
    realiCloneHash = newHash(12);
    sprintf(query, 
    	"select * from cloneAliPos where chrom='%s'and chromStart<%u and chromEnd>%u",
	chromName, winEnd, winStart);
    sr = sqlGetResult(conn, query);
    while ((row = sqlNextRow(sr)) != NULL)
	{
	clonePosStaticLoad(row, &cp);
	AllocVar(ci);
	hel = hashAdd(realiCloneHash, cp.name, ci);
	ci->name = hel->name;
	ci->cloneStart = cp.chromStart;
	ci->cloneEnd = cp.chromEnd;
	ci->phase = cp.phase;
	slAddHead(&realiCloneList, ci);
	}
    sqlFreeResult(&sr);

    /* Load in alignments from database and sort them by clone. */
    sprintf(query, "select * from %s_frags where tStart<%u and tEnd>%u",
	chromName, winEnd, winStart);
    sr = sqlGetResult(conn, query);
    while ((row = sqlNextRow(sr)) != NULL)
	{
	psl = pslLoad(row);
	fragName = psl->qName;
	strcpy(cloneName, fragName);
	s = strchr(cloneName, '_');
	if (s != NULL)
	    *s = 0;
	if ((hel = hashLookup(realiCloneHash, cloneName)) == NULL)
	    {
	    warn("%s not in range in cloneAliPos");
	    continue;
	    }
	ci = hel->val;
	if ((cf = findCloneFrag(ci, fragName)) == NULL)
	    {
	    AllocVar(cf);
	    cf->name = fragName;
	    slAddHead(&ci->fragList, cf);
	    }
	AllocVar(cfa);
	cfa->frag = cf;
	cfa->psl = psl;
	cfa->start = psl->tStart;
	cfa->end = psl->tEnd;
	slAddHead(&ci->cfaList, cfa);
	}
    slSort(&realiCloneList, cmpCloneInfo);
    sqlFreeResult(&sr);
    hFreeConn(&conn);

    /* Do preliminary layout processing for each clone. */
    for (ci = realiCloneList; ci != NULL; ci = ci->next)
	{
	slReverse(&ci->cfaList);
	layoutCloneAli(ci);
	}
    }
}


struct hash *glCloneHash;
struct cloneInfo *glCloneList;

void glLoadInWindow()
/* Load up glCloneHash and glCloneList with the clones in the window. */
{
if (glCloneList == NULL)
    {
    char query[256];
    struct sqlConnection *conn = hAllocConn();
    struct sqlResult *sr = NULL;
    char **row;
    struct cloneInfo *ci;
    struct gl *gl;
    char *fragName;
    struct cloneFrag *cf;
    char cloneName[128];
    struct hashEl *hel;
    struct cloneFragPos *cfa;
    struct clonePos cp;
    char *s;

    /* Load in clone extents from database. */
    glCloneHash = newHash(12);
    sprintf(query, 
    	"select * from clonePos where chrom='%s'and chromStart<%u and chromEnd>%u",
	chromName, winEnd, winStart);
    sr = sqlGetResult(conn, query);
    while ((row = sqlNextRow(sr)) != NULL)
	{
	clonePosStaticLoad(row, &cp);
	AllocVar(ci);
	hel = hashAdd(glCloneHash, cp.name, ci);
	ci->name = hel->name;
	ci->cloneStart = cp.chromStart;
	ci->cloneEnd = cp.chromEnd;
	ci->phase = cp.phase;
	ci->stage = cp.stage[0];
	slAddHead(&glCloneList, ci);
	}
    sqlFreeResult(&sr);

    /* Load in alignments from database and sort them by clone. */
    sprintf(query, "select * from %s_gl where start<%u and end>%u",
	chromName, winEnd, winStart);
    sr = sqlGetResult(conn, query);
    while ((row = sqlNextRow(sr)) != NULL)
	{
	gl = glLoad(row);
	fragName = gl->frag;
	strcpy(cloneName, fragName);
	s = strchr(cloneName, '_');
	if (s != NULL)
	    *s = 0;
	if ((hel = hashLookup(glCloneHash, cloneName)) == NULL)
	    {
	    warn("%s not in range in clonePos", cloneName);
	    continue;
	    }
	ci = hel->val;
	if ((cf = findCloneFrag(ci, fragName)) == NULL)
	    {
	    AllocVar(cf);
	    cf->name = fragName;
	    slAddHead(&ci->fragList, cf);
	    }
	AllocVar(cfa);
	cfa->frag = cf;
	cfa->gl = gl;
	cfa->start = gl->start;
	cfa->end = gl->end;
	slAddHead(&ci->cfaList, cfa);
	}
    slSort(&glCloneList, cmpCloneInfo);
    sqlFreeResult(&sr);
    hFreeConn(&conn);

    /* Do preliminary layout processing for each clone. */
    for (ci = glCloneList; ci != NULL; ci = ci->next)
	{
	slReverse(&ci->cfaList);
	layoutCloneAli(ci);
	}
    }
}



void coverageLoad(struct trackGroup *tg)
/* Load up clone alignments from database tables and organize. */
{
glLoadInWindow();
tg->items = glCloneList;
}

void coverageFree(struct trackGroup *tg)
/* Free up clone track group items. */
{
cloneInfoFreeList(&glCloneList);
freeHash(&glCloneHash);
}

struct trackGroup *coverageTrackGroup()
/* Make track group for golden path positions of all frags. */
{
struct trackGroup *tg;

AllocVar(tg);
tg->mapName = "hgCover";
tg->visibility = tvDense;
tg->longLabel = "Clone Coverage/Fragment Position";
tg->shortLabel = "Coverage";
tg->loadItems = coverageLoad;
tg->freeItems = coverageFree;
tg->drawItems = cloneDraw;
tg->altColor.r = 180;
tg->altColor.g = 180;
tg->altColor.b = 180;
tg->itemName = cloneName;
tg->mapItemName = cloneName;
tg->totalHeight = cloneTotalHeight;
tg->itemHeight = cloneItemHeight;
return tg;
}


void gapLoad(struct trackGroup *tg)
/* Load up clone alignments from database tables and organize. */
{
bedLoadChrom(tg, "gap", (ItemLoader)agpGapLoad);
}

void gapFree(struct trackGroup *tg)
/* Free up isochore items. */
{
agpGapFreeList((struct agpGap**)&tg->items);
}

char *gapName(struct trackGroup *tg, void *item)
/* Return name of gold track item. */
{
static char buf[24];
struct agpGap *gap = item;
sprintf(buf, "%s %s", gap->type, gap->bridge);
return buf;
}

static void gapDraw(struct trackGroup *tg, int seqStart, int seqEnd,
        struct memGfx *mg, int xOff, int yOff, int width, 
        MgFont *font, Color color, enum trackVisibility vis)
/* Draw gap items. */
{
int baseWidth = seqEnd - seqStart;
struct agpGap *item;
int y = yOff;
int heightPer = tg->heightPer;
int lineHeight = tg->lineHeight;
int x1,x2,w;
boolean isFull = (vis == tvFull);
double scale = width/(double)baseWidth;
int halfSize = heightPer/2;

for (item = tg->items; item != NULL; item = item->next)
    {
    x1 = round((double)((int)item->chromStart-winStart)*scale) + xOff;
    x2 = round((double)((int)item->chromEnd-winStart)*scale) + xOff;
    w = x2-x1;
    if (w < 1)
	w = 1;
    if (sameString(item->bridge, "no"))
	mgDrawBox(mg, x1, y, w, heightPer, color);
    else  /* Leave white line in middle of bridged gaps. */
        {
	mgDrawBox(mg, x1, y, w, halfSize, color);
	mgDrawBox(mg, x1, y+heightPer-halfSize, w, halfSize, color);
	}
    if (isFull)
	{
	char name[32];
	sprintf(name, "%s", item->type);
	printf("<AREA SHAPE=RECT COORDS=\"%d,%d,%d,%d\" ", x1, y, x1+w, y+heightPer);
	printf("HREF=\"../cgi-bin/hgc?g=%s&i=%s&c=%s&l=%d&r=%d&o=%d\" ", 
	    tg->mapName, name, chromName, winStart, winEnd, item->chromStart);
	printf("ALT= \"%s bridged? %s\" TITLE=\"%s bridged? %s\">\n", item->type, item->bridge,
		item->type, item->bridge); 
	y += lineHeight;
	}
    }
}


struct trackGroup *gapTg()
/* Make track group for golden path positions of all frags. */
{
struct trackGroup *tg;
AllocVar(tg);
tg->mapName = "hgGap";
tg->visibility = tvDense;
tg->longLabel = "Gap Locations";
tg->shortLabel = "Gap";
tg->loadItems = gapLoad;
tg->freeItems = gapFree;
tg->drawItems = gapDraw;
tg->itemName = gapName;
tg->mapItemName = gapName;
tg->totalHeight = tgFixedTotalHeight;
tg->itemHeight = tgFixedItemHeight;
return tg;
}


#ifdef SOMEDAY
/* These next three vars are used to communicate info from the
 * loadAltGraph routine to the agTransOut routine */
struct linkedFeatures *agTempList;  /* Transcripts get added to this list. */
int agTempOrientation;
int agClusterIx;
int agTranscriptIx;

void agTransOut(struct ggAliInfo *da, int cStart, int cEnd)
/* Called to convert gene graph transcript to something more permanent. */
{
struct linkedFeatures *lf;
struct simpleFeature *sfList = NULL, *sf;
struct ggVertex *vertices = da->vertices;
int vertexCount = da->vertexCount;
int i;
static int tot = 0;
int minStart = 0x3fffffff;
int maxEnd = -minStart;
int start,end;

AllocVar(lf);
lf->grayIx = maxShade;
sprintf(lf->name, "cluster %d.%d", agClusterIx+1, agTranscriptIx++);
lf->orientation = agTempOrientation;
for (i=0; i<vertexCount; i+=2)
    {
    AllocVar(sf);
    start = sf->start = vertices[i].position;
    if (start < minStart)
	minStart = start;
    end = sf->end = vertices[i+1].position;
    if (end > maxEnd)
	maxEnd = end;
    sf->grayIx = maxShade;
    slAddHead(&sfList, sf);
    }
lf->start = minStart;
lf->end = maxEnd;
slReverse(&sfList);
lf->components = sfList;
slAddHead(&agTempList, lf);
}

void loadAltGraph(struct trackGroup *tg)
/* Get alt splicing graphs and turn them into linked features. */
{
struct hgBac *bac = hgGetBac(chromName);
char query[256];
struct sqlConnection *conn = hgAllocConn();
struct sqlResult *sr = NULL;
char **row;
boolean firstTime = TRUE;
boolean anyIntersection = FALSE;
HGID lastParent = 0;

sprintf(query, 
    "select * from altGraph where startBac=%u and endPos>%d and startPos<%d",
    bac->id, winStart, winEnd);

sr = sqlGetResult(conn, query);
agTempList = NULL;
agClusterIx = 0;
while ((row = sqlNextRow(sr)) != NULL)
    {
    struct geneGraph *gg = ggFromRow(row);
    agTranscriptIx = 0;
    agTempOrientation = gg->orientation;
    traverseGeneGraph(gg, gg->startPos, gg->endPos, agTransOut);
    freeGeneGraph(&gg);
    ++agClusterIx;
    }
slReverse(&agTempList);
tg->items = agTempList;
agTempList = NULL;
sqlFreeResult(&sr);
hgFreeConn(&conn);
}

struct trackGroup *altGraphTg()
/* Make track group of altGraph. */
{
struct trackGroup *tg = linkedFeaturesTg();
tg->mapName = "hgAltGraph";
tg->visibility = tvFull;
tg->longLabel = "mRNA Clusters";
tg->shortLabel = "mRNA Clusters";
tg->loadItems = loadAltGraph;
return tg;
}
#endif /* SOMEDAY */

enum trackVisibility limitVisibility(struct trackGroup *tg)
/* Return default visibility limited by number of items. */
{
enum trackVisibility vis = tg->visibility;
if (vis != tvFull)
    return vis;
if (slCount(tg->items) <= maxItemsInFullTrack)
    return tvFull;
else
    return tvDense;
}

void makeActiveImage(struct trackGroup *groupList)
/* Make image and image map. */
{
struct trackGroup *group;
MgFont *font = tl.font;
struct memGfx *mg;
struct tempName gifTn;
char *mapName = "map";

int fontHeight = mgFontLineHeight(font);
int insideHeight = fontHeight-1;
int border = 1;
int xOff = border;
int pixWidth, pixHeight;
int insideWidth;
int y;

/* Coordinates of open/close/hide buttons. */
#ifdef SOON
int openCloseHideWidth = 150;
#endif
int openCloseHideWidth = 0;
int openWidth = 50, openOffset = 0;
int closeWidth = 50, closeOffset = 50;
int hideWidth = 50, hideOffset = 100;

int typeCount = slCount(groupList);
int leftLabelWidth = 0;

int rulerHeight = fontHeight;
int relNumOff;
int i;

/* Figure out dimensions and allocate drawing space. */
if (withLeftLabels)
    {
    leftLabelWidth = tl.leftLabelWidth;
    xOff += leftLabelWidth + border;
    }    
pixWidth = tl.picWidth;
insideWidth = pixWidth-border-xOff;
pixHeight = border + rulerHeight;
for (group = groupList; group != NULL; group = group->next)
    {
    if (group->visibility != tvHide)
	{
	pixHeight += group->totalHeight(group, group->limitedVis);
	if (withCenterLabels)
	    pixHeight += fontHeight;
	}
    }
mg = mgNew(pixWidth, pixHeight);
mgClearPixels(mg);
makeGrayShades(mg);

/* Start up client side map. */
printf("<MAP Name=%s>\n", mapName);
/* Find colors to draw in. */
for (group = groupList; group != NULL; group = group->next)
    {
    if (group->limitedVis != tvHide)
	{
	group->ixColor = mgFindColor(mg, 
		group->color.r, group->color.g, group->color.b);
	group->ixAltColor = mgFindColor(mg, 
		group->altColor.r, group->altColor.g, group->altColor.b);
	}
    }

/* Draw left labels. */
if (withLeftLabels)
    {
    int inWid = xOff-border*3;
    mgDrawBox(mg, xOff-border*2, 0, border, pixHeight, mgFindColor(mg, 0, 0, 200));
    mgSetClip(mg, border, border, inWid, pixHeight-2*border);
    y = border;
    for (group = groupList; group != NULL; group = group->next)
        {
	struct slList *item;
	switch (group->limitedVis)
	    {
	    case tvHide:
		break;	/* Do nothing; */
	    case tvFull:
		if (withCenterLabels)
		    y += fontHeight;
		for (item = group->items; item != NULL; item = item->next)
		    {
		    char *name = group->itemName(group, item);
		    int itemHeight = group->itemHeight(group, item);
		    mgTextCentered(mg, border, y, inWid, itemHeight, group->ixColor, font, name);
		    y += itemHeight;
		    }
		break;
	    case tvDense:
		if (withCenterLabels)
		    y += fontHeight;
		mgTextCentered(mg, border, y, inWid, group->lineHeight, group->ixColor, font, group->shortLabel);
		y += group->lineHeight;
		break;
	    }
        }
    }

/* Draw guidelines. */
if (withGuidelines)
    {
    int clWidth = insideWidth-openCloseHideWidth;
    int ochXoff = xOff + clWidth;
    int height = pixHeight - 2*border;
    int x;
    Color color = mgFindColor(mg, 220, 220, 255);

    mgSetClip(mg, xOff, border, insideWidth, height);
    y = border;

    for (x = xOff+9; x<pixWidth; x += 10)
	mgDrawBox(mg, x, y, 1, height, color);
    }

/* Draw center labels. */
if (withCenterLabels)
    {
    int clWidth = insideWidth-openCloseHideWidth;
    int ochXoff = xOff + clWidth;
    mgSetClip(mg, xOff, border, insideWidth, pixHeight - 2*border);
    y = border;
    for (group = groupList; group != NULL; group = group->next)
        {
	if (group->limitedVis != tvHide)
	    {
	    Color color = group->ixColor;
	    mgTextCentered(mg, xOff, y+1, clWidth, insideHeight, color, font, group->longLabel);
	    mapBoxToggleVis(0,y+1,pixWidth,insideHeight,group);
#ifdef SOON
	    uglyf("ochXoff = %d, ochWidth = %d, pixWidth = %d", ochXoff, openCloseHideWidth, pixWidth);
	    mgTextCentered(mg, ochXoff, y+1, openCloseHideWidth, insideHeight, 
		color, font, "[open] [close] [hide]");
#endif
	    y += fontHeight;
	    y += group->height;
	    }
        }
    }

/* Draw tracks. */
y = 1;
for (group = groupList; group != NULL; group = group->next)
    {
    if (group->limitedVis != tvHide)
	{
	if (withCenterLabels)
	    y += fontHeight;
	mgSetClip(mg, xOff, y, insideWidth, group->height);
	group->drawItems(group, winStart, winEnd,
	    mg, xOff, y, insideWidth, 
	    font, group->ixColor, group->limitedVis);
	y += group->height;
	}
    }

/* Show ruler at bottom. */
y = pixHeight - rulerHeight;
mgSetClip(mg, xOff, y, insideWidth, rulerHeight);
relNumOff = winStart;
if (seqReverse)
    relNumOff = -relNumOff;
mgDrawRuler(mg, xOff, y, rulerHeight, insideWidth, MG_BLACK, font,
    relNumOff, winBaseCount);
mapBoxHc(xOff,y,insideWidth,rulerHeight, "getDna", "mixed", "Fetch DNA");

/* Make map background. */
    {
    y = border;
    for (group = groupList; group != NULL; group = group->next)
        {
	struct slList *item;
	switch (group->limitedVis)
	    {
	    case tvHide:
		break;	/* Do nothing; */
	    case tvFull:
		if (withCenterLabels)
		    y += fontHeight;
		for (item = group->items; item != NULL; item = item->next)
		    {
		    int height = group->itemHeight(group, item);
		    if (group->itemMap)
		        group->itemMap(group, item, 0, y, pixWidth, height);
		    else
			mapBoxHc(0,y,pixWidth,height, group->mapName, 
			    group->mapItemName(group, item), 
			    group->itemName(group, item));
		    y += height;
		    }
		break;
	    case tvDense:
		if (withCenterLabels)
		    y += fontHeight;
		mapBoxToggleVis(0,y,pixWidth,group->lineHeight,group);
		y += group->lineHeight;
		break;
	    }
        }
    }

/* Finish map. */
printf("</MAP>\n");

/* Save out picture and tell html file about it. */
makeTempName(&gifTn, "hgt", ".gif");
mgSaveGif(mg, gifTn.forCgi);
printf(
    "<P><IMG SRC = \"%s\" BORDER=1 WIDTH=%d HEIGHT=%d USEMAP=#%s><BR>\n",
    gifTn.forHtml, pixWidth, pixHeight, mapName);

mgFree(&mg);
}


void makeNumText(char *name, int num, int digits)
/* Make a text control filled with a number. */
{
printf("<INPUT TYPE=TEXT NAME=\"%s\" SIZE=%d VALUE=%d>", name, digits, num);
}

void makeText(char *name, char *initialVal, int chars)
/* Make a text control filled with initial value. */
{
printf("<INPUT TYPE=TEXT NAME=\"%s\" SIZE=%d VALUE=%s>", name, chars, initialVal);
}

void makeSubmitButton(char *name, char *value)
/* Make a submit button. */
{
printf("<INPUT TYPE=SUBMIT NAME=\"%s\" VALUE=\"%s\">", name, value);
}

void makeDropList(char *name, char *menu[], int menuSize, char *checked)
/* Make a drop-down list. */
{
int i;
char *selString;
if (checked == NULL) checked = menu[0];
printf("<SELECT ALIGN=CENTER NAME=\"%s\">", name);
for (i=0; i<menuSize; ++i)
    {
    if (!differentWord(menu[i], checked))
        selString = " SELECTED";
    else
        selString = "";
    printf("<OPTION%s>%s</OPTION>", selString, menu[i]);
    }
printf("</SELECT>");
}

void makeCheckBox(char *name, boolean isChecked)
/* Create a checkbox with the given name in the given state. */
{
printf("<INPUT TYPE=CHECKBOX NAME=\"%s\" VALUE=on%s>", name,
    (isChecked ? " CHECKED" : "") );
}

void doForm()
/* Make the tracks display form with the zoom/scroll
 * buttons and the active image. */
{
struct trackGroup *group;

if (calledSelf)
    {
    /* Set center and left labels from UI. */
    withLeftLabels = cgiBoolean("leftLabels");
    withCenterLabels = cgiBoolean("centerLabels");
    withGuidelines = cgiBoolean("guidelines");
    }

/* Make list of all track groups. */
slSafeAddHead(&tGroupList, cytoBandTg());
slSafeAddHead(&tGroupList, genethonTg());
slSafeAddHead(&tGroupList, isochoresTg());
slSafeAddHead(&tGroupList, gcPercentTg());
slSafeAddHead(&tGroupList, contigTg());
slSafeAddHead(&tGroupList, goldTrackGroup());
slSafeAddHead(&tGroupList, gapTg());
if (tableExists("genomicDups"))
    slSafeAddHead(&tGroupList, genomicDupsTg());
#ifdef SOON
#endif /* SOON */
slSafeAddHead(&tGroupList, coverageTrackGroup());
slSafeAddHead(&tGroupList, genieKnownTg());
slSafeAddHead(&tGroupList, genieAltTg());
slSafeAddHead(&tGroupList, ensemblGeneTg());
slSafeAddHead(&tGroupList, fullMrnaTg());
slSafeAddHead(&tGroupList, estTg());
if (tableExists("est3"))
    slSafeAddHead(&tGroupList, est3Tg());
slSafeAddHead(&tGroupList, cpgIslandTg());
if (chromTableExists("_tet_waba"))
   slSafeAddHead(&tGroupList, tetTg());
slSafeAddHead(&tGroupList, repeatTg());
slSafeAddHead(&tGroupList, simpleRepeatTg());
slReverse(&tGroupList);

/* Get visibility values if any from ui. */
for (group = tGroupList; group != NULL; group = group->next)
    {
    char *s = cgiOptionalString(group->mapName);
    if (s != NULL)
	{
	int vis = stringArrayIx(s, tvStrings, ArraySize(tvStrings));
	if (vis < 0)
	    errAbort("Unknown value %s for %s", s, group->mapName);
	group->visibility = vis;
	}
    }

/* Tell groups to load their items. */
for (group = tGroupList; group != NULL; group = group->next)
    {
    if (group->visibility != tvHide)
	{
	group->loadItems(group); 
	group->limitedVis = limitVisibility(group);
	}
    }

/* Tell browser where to go when they click on image. */
printf("<FORM ACTION=\"%shgTracks\">\n\n", cgiDir());

/* Center everything from now on. */
printf("<CENTER>\n");

/* Show title . */
printf("<H2>Chromosome %s, Bases %d-%d</H2>", skipChr(chromName), winStart+1, winEnd);

/* Put up scroll and zoom controls. */
fputs("move ", stdout);
makeSubmitButton("left3", "<<<");
makeSubmitButton("left2", " <<");
makeSubmitButton("left1", " < ");
makeSubmitButton("right1", " > ");
makeSubmitButton("right2", ">> ");
makeSubmitButton("right3", ">>>");
fputs(" zoom in ", stdout);
makeSubmitButton("in1", "1.5x");
makeSubmitButton("in2", " 3x ");
makeSubmitButton("in3", "10x");
fputs(" zoom out ", stdout);
makeSubmitButton("out1", "1.5x");
makeSubmitButton("out2", " 3x ");
makeSubmitButton("out3", "10x");
fputs("<BR>\n", stdout);

/* Make line that says position. */
    {
    fputs("position ", stdout);
    makeText("position", position, 30);
    makeSubmitButton("submit", "jump");
    printf(" <A HREF=\"../goldenPath/help/hgTracksHelp.html\" TARGET=_blank>User's Guide</A>");
    fputs("<BR>\n", stdout);
    }

/* Make clickable image and map. */
makeActiveImage(tGroupList);
fputs("Click on an item in a track to view more information on that item. "
      "Click center label to toggle between full and dense display of "
      "that track.  Tracks with more than 300 items are always displayed "
      "densely.  Click on ruler for DNA sequence.", stdout);

/* Display bottom control panel. */
htmlHorizontalLine();
fputs("Chromosome ", stdout);
makeDropList("seqName", chromNames, 24, chromName);
fputs(" bases ",stdout);
makeNumText("winStart", winStart, 12);
fputs(" - ", stdout);
makeNumText("winEnd", winEnd, 12);
printf(" Guidelines ");
makeCheckBox("guidelines", withGuidelines);
printf(" <B>Labels:</B> ");
printf("left ");
makeCheckBox("leftLabels", withLeftLabels);
printf("center ");
makeCheckBox("centerLabels", withCenterLabels);
printf(" ");
makeSubmitButton("submit", "refresh");
printf("<BR>\n");


/* Display viewing options for each group. */
printf("<B>Track Controls:</B> ");
for (group = tGroupList; group != NULL; group = group->next)
    {
    printf(" %s ", group->shortLabel);
    makeDropList(group->mapName, tvStrings, ArraySize(tvStrings), tvStrings[group->visibility]);
    }
saveHiddenVars();

/* Clean up. */
for (group = tGroupList; group != NULL; group = group->next)
    {
    if (group->visibility != tvHide)
	group->freeItems(group);
    }
printf("</FORM>");
}


void zoomAroundCenter(double amount)
/* Set ends so as to zoom around center by scaling amount. */
{
int center = (winStart + winEnd)/2;
int newCount = (int)(winBaseCount*amount + 0.5);
if (newCount < 30) newCount = 30;
if (newCount > seqBaseCount)
    newCount = seqBaseCount;
winStart = center - newCount/2;
winEnd = winStart + newCount;
if (winStart <= 0 && winEnd >= seqBaseCount)
    {
    winStart = 0;
    winEnd = seqBaseCount;
    }
else if (winStart <= 0)
    {
    winStart = 0;
    winEnd = newCount;
    }
else if (winEnd > seqBaseCount)
    {
    winEnd = seqBaseCount;
    winStart = winEnd - newCount;
    }
winBaseCount = winEnd - winStart;
}

void relativeScroll(double amount)
/* Scroll percentage of visible window. */
{
int offset;
int newStart, newEnd;

if (seqReverse)
    amount = -amount;
offset = (int)(amount * winBaseCount + 0.5);
/* Make sure don't scroll of ends. */
newStart = winStart + offset;
newEnd = winEnd + offset;
if (newStart < 0)
    offset = -winStart;
else if (newEnd > seqBaseCount)
    offset = seqBaseCount - winEnd;

/* Move window. */
winStart += offset;
winEnd += offset;
}


char *isHumanChrom(char *chrom)
/* Returns "cannonical" name of chromosome or NULL
 * if not a chromosome. */
{
int i;
for (i=0; i<chromCount; ++i)
    if (sameWord(chromNames[i], chrom))
	return chromNames[i];
return NULL;
}

char *startsWithShortHumanChromName(char *chrom)
/* Return "cannonical" name of chromosome or NULL
 * if not a chromosome.  This expects no 'chr' in name. */
{
int num;
char buf[64];
char c = chrom[0];

if (c == 'x' || c == 'X' || c == 'Y' || c == 'y')
    {
    sprintf(buf, "chr%c", toupper(c));
    return isHumanChrom(buf);
    }
if (!isdigit(chrom[0]))
    return NULL;
num = atoi(chrom);
if (num < 1 || num > 22)
    return NULL;
sprintf(buf, "chr%d", num);
return isHumanChrom(buf);
}


boolean parseHumanChromRange(char *spec, char **retChromName, 
	int *retWinStart, int *retWinEnd)
/* Parse something of form chrom:start-end into pieces. */
{
char *chrom, *start, *end;
char buf[256];
strncpy(buf, spec, sizeof(buf));
chrom = buf;
start = strchr(chrom, ':');
if (start == NULL)
    return FALSE;
else 
    *start++ = 0;
end = strchr(start, '-');
if (end == NULL)
    return FALSE;
else
    *end++ = 0;
chrom = trimSpaces(chrom);
start = trimSpaces(start);
end = trimSpaces(end);
if (!isdigit(start[0]))
    return FALSE;
if (!isdigit(end[0]))
    return FALSE;
if ((chrom = isHumanChrom(chrom)) == NULL)
    return FALSE;
if (retChromName != NULL)
    *retChromName = chrom;
if (retWinStart != NULL)
    *retWinStart = atoi(start)-1;
if (retWinEnd != NULL)
    *retWinEnd = atoi(end);
return TRUE;
}

boolean isHumanChromRange(char *spec)
/* Returns TRUE if spec is chrom:N-M for some human
 * chromosome chrom and some N and M. */
{
return parseHumanChromRange(spec, NULL, NULL, NULL);
}

void findContigPos(char *contig, char **retChromName, 
	int *retWinStart, int *retWinEnd)
/* Find position in genome of contig.  Don't alter
 * return variables if some sort of error. */
{
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr = NULL;
struct dyString *query = newDyString(256);
char **row;
struct ctgPos *ctgPos;
dyStringPrintf(query, "select * from ctgPos where contig = '%s'", contig);
sr = sqlMustGetResult(conn, query->string);
row = sqlNextRow(sr);
if (row == NULL)
    errAbort("Couldn't find contig %s", contig);
ctgPos = ctgPosLoad(row);
*retChromName = isHumanChrom(ctgPos->chrom);
*retWinStart = ctgPos->chromStart;
*retWinEnd = ctgPos->chromEnd;
ctgPosFree(&ctgPos);
freeDyString(&query);
sqlFreeResult(&sr);
hFreeConn(&conn);
}

boolean isContigName(char *contig)
/* REturn TRUE if a FPC contig name. */
{
return startsWith("ctg", contig);
}

boolean isAccForm(char *s)
/* Returns TRUE if s is of format to be a genbank accession. */
{
int len = strlen(s);
if (len < 6 || len > 10)
    return FALSE;
if (!isalpha(s[0]))
    return FALSE;
if (!isdigit(s[len-1]))
    return FALSE;
return TRUE;
}

boolean findClonePos(char *spec, char **retChromName, 
	int *retWinStart, int *retWinEnd)
/* Return clone position. */
{
if (!isAccForm(spec))
    return FALSE;
else
    {
    struct sqlConnection *conn = hAllocConn();
    struct sqlResult *sr = NULL;
    struct dyString *query = newDyString(256);
    char **row;
    boolean ok = FALSE;
    struct clonePos *clonePos;
    dyStringPrintf(query, "select * from clonePos where name like '%s%%'", spec);
    sr = sqlMustGetResult(conn, query->string);
    row = sqlNextRow(sr);
    if (row != NULL)
	{
	clonePos = clonePosLoad(row);
	*retChromName = isHumanChrom(clonePos->chrom);
	*retWinStart = clonePos->chromStart;
	*retWinEnd = clonePos->chromEnd;
	clonePosFree(&clonePos);
	ok = TRUE;
	}
    freeDyString(&query);
    sqlFreeResult(&sr);
    hFreeConn(&conn);
    return ok;
    }
}

boolean findCytoBand(char *spec, char **retChromName, int *retWinStart, int *retWinEnd)
/* Return position associated with cytological band if spec looks to be in that form. */
{
char *fullChromName, *shortChromName;
int len;
char c;

/* First make sure spec is in format to be a band name. */
if ((fullChromName = startsWithShortHumanChromName(spec)) == NULL)
    return FALSE;
shortChromName = skipChr(fullChromName);
len = strlen(shortChromName);
spec += len;
c = spec[0];
if (c != 'p' && c != 'q')
    return FALSE;
if (!isdigit(spec[1]))
    return FALSE;

/* Have parsed out spec and it looks like a cytological band.  Look up 
 * band in database now. */
    {
    struct sqlConnection *conn = hAllocConn();
    struct sqlResult *sr = NULL;
    struct dyString *query = newDyString(256);
    char **row;
    boolean ok = FALSE;
    struct cytoBand *cytoBand;
    dyStringPrintf(query, "select * from cytoBand where chrom = '%s' and name = '%s'", 
    	fullChromName, spec);
    sr = sqlMustGetResult(conn, query->string);
    row = sqlNextRow(sr);
    if (row == NULL)
        {
	sqlFreeResult(&sr);
	dyStringClear(query);
	dyStringPrintf(query, "select * from cytoBand where chrom = '%s' and name like '%s%%'", 
	    fullChromName, spec);
	sr = sqlMustGetResult(conn, query->string);
	row = sqlNextRow(sr);
	}
    if (row != NULL)
	{
	cytoBand = cytoBandLoad(row);
	*retChromName = fullChromName;
	*retWinStart = cytoBand->chromStart;
	*retWinEnd = cytoBand->chromEnd;
	cytoBandFree(&cytoBand);
	ok = TRUE;
	}
    freeDyString(&query);
    sqlFreeResult(&sr);
    hFreeConn(&conn);
    return ok;
    }
}

boolean findStsPos(char *table, char *spec, 
   char **retChromName, int *retWinStart, int *retWinEnd)
/* Look for position in some STS table. */
{
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr = NULL;
struct dyString *query = newDyString(256);
char **row;
boolean ok = FALSE;
struct mapSts *mapSts;

dyStringPrintf(query, "select * from %s where name = '%s'", table, spec);
sr = sqlMustGetResult(conn, query->string);
row = sqlNextRow(sr);
if (row != NULL)
    {
    mapSts = mapStsLoad(row);
    *retChromName = isHumanChrom(mapSts->chrom);
    *retWinStart = mapSts->chromStart - 100000;
    *retWinEnd = mapSts->chromEnd + 100000;
    mapStsFree(&mapSts);
    ok = TRUE;
    }
freeDyString(&query);
sqlFreeResult(&sr);
hFreeConn(&conn);
return ok;
}

boolean findGenethonPos(char *spec, char **retChromName, int *retWinStart, int *retWinEnd)
/* See if it's a genethon map position. */
{
return findStsPos("mapGenethon", spec, retChromName, retWinStart, retWinEnd);
}

char *mrnaType(char *acc)
/* Return "mrna" or "est" if acc is mRNA, otherwise NULL. */
{
static char typeBuf[16];
char *type;
struct sqlConnection *conn = hAllocConn();
char query[128];
sprintf(query, "select type from mrna where acc = '%s'", acc);
type = sqlQuickQuery(conn, query, typeBuf, sizeof(typeBuf));
hFreeConn(&conn);
return type;
}

struct psl *findAllAli(char *acc, char *type)
/* Find all alignments of the given type. */
{
struct sqlConnection *conn = hAllocConn();
char query[256];
struct psl *pslList = NULL, *psl;
struct sqlResult *sr;
char **row;

sprintf(query, "select * from all_%s where qName = '%s'", type, acc);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    psl = pslLoad(row);
    slAddHead(&pslList, psl);
    }
hFreeConn(&conn);
slReverse(&pslList);
return pslList;
}

int pslMrnaScore(const struct psl *psl)
/* Return a simple score for psl. */
{
return psl->match + (psl->repMatch>>1) - psl->misMatch - psl->qNumInsert;
}

int pslCmpScore(const void *va, const void *vb)
/* Compare two psl to sort by score position . */
{
const struct psl *a = *((struct psl **)va);
const struct psl *b = *((struct psl **)vb);
return pslMrnaScore(b) - pslMrnaScore(a);
}

struct slName *accsThatMatchKey(char *key, char *field)
/* Return list of accessions that match key on a particular field. */
{
struct slName *list = NULL, *el;
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr;
char **row;
char query[256];

sprintf(query, "select mrna.acc from mrna,%s where %s.name like '%%%s%%'  and mrna.%s = %s.id", field, field, key, field, field);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    el = newSlName(row[0]);
    slAddHead(&list, el);
    }
sqlFreeResult(&sr);
hFreeConn(&conn);
slReverse(&list);
return list;
}

void findHitsToTables(char *key, char *tables[], int tableCount, 
    struct hash **retHash, struct slName **retList)
/* Return all unique accessions that match any table. */
{
struct slName *list = NULL, *el;
struct hash *hash = newHash(0);
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr;
char **row;
char query[256];
char *field;
int i;

for (i = 0; i<tableCount; ++i)
    {
    struct slName *idList = NULL, *idEl;
    
    /* I'm doing this query in two steps in C rather than
     * in one step in SQL just because it somehow is much
     * faster this way (like 100x faster) in mySQL. */
    field = tables[i];
    sprintf(query, "select id from %s where name like '%%%s%%'", 
       field, key);
    sr = sqlGetResult(conn, query);
    while ((row = sqlNextRow(sr)) != NULL)
	{
	idEl = newSlName(row[0]);
	slAddHead(&idList, idEl);
	}
    sqlFreeResult(&sr);
    for (idEl = idList; idEl != NULL; idEl = idEl->next)
        {
	sprintf(query, "select acc from mrna where %s = %s", field, idEl->name);
	sr = sqlGetResult(conn, query);
	while ((row = sqlNextRow(sr)) != NULL)
	    {
	    char *acc = row[0];
	    if (!hashLookup(hash, acc))
		{
		el = newSlName(acc);
		slAddHead(&list, el);
		hashAdd(hash, acc, NULL);
		}
	    }
	sqlFreeResult(&sr);
        }
    slFreeList(&idList);
    }
hFreeConn(&conn);
slReverse(&list);
*retList = list;
*retHash = hash;
}


void andHits(struct hash *aHash, struct slName *aList, 
	struct hash *bHash, struct slName *bList,
	struct hash **retHash, struct slName **retList)
/* Return hash/list that is intersection of lists a and b. */
{
struct slName *list = NULL, *el, *newEl;
struct hash *hash = newHash(0);
for (el = aList; el != NULL; el = el->next)
    {
    char *name = el->name;
    if (hashLookup(bHash, name) && !hashLookup(hash, name))
        {
	newEl = newSlName(name);
	slAddHead(&list, newEl);
	hashAdd(hash, name, NULL);
	}
    }
*retHash = hash;
*retList = list;
}

boolean findMrnaKeys(char *keys)
/* Find mRNA that has keyword in one of it's fields. */
{
char *words[32];
char buf[512];
int wordCount;
struct slName *el;
static char *tables[] = {
	"productName", "geneName",
	"author", "tissue", "cell", "description", "development", 
	};
struct hash *oneKeyHash = NULL;
struct slName *oneKeyList = NULL;
struct hash *allKeysHash = NULL;
struct slName *allKeysList = NULL;
struct hash *andedHash = NULL;
struct slName *andedList = NULL;
int i;

strncpy(buf, keys, sizeof(buf));
wordCount = chopLine(buf, words);
if (wordCount == 0)
    return FALSE;
for (i=0; i<wordCount; ++i)
    {
    findHitsToTables(words[i], tables, ArraySize(tables), &oneKeyHash, &oneKeyList);
    if (allKeysHash == NULL)
        {
	allKeysHash = oneKeyHash;
	oneKeyHash = NULL;
	allKeysList = oneKeyList;
	oneKeyList = NULL;
	}
    else
        {
	andHits(oneKeyHash, oneKeyList, allKeysHash, allKeysList, &andedHash, &andedList);
	freeHash(&oneKeyHash);
	slFreeList(&oneKeyList);
	freeHash(&allKeysHash);
	slFreeList(&allKeysList);
	allKeysHash = andedHash;
	andedHash = NULL;
	allKeysList = andedList;
	andedList = NULL;
	}
    }
if (allKeysList == NULL)
    return FALSE;
printf("<H2>Text Search Results</H2>\n");
printf("The following mRNA accessions match  your search criteria<BR>\n");
/* Fetch descriptions of all matchers and display. */
    {
    struct sqlConnection *conn = hAllocConn();
    struct sqlResult *sr;
    char **row;
    char query[256];
    char description[512];
    for (el = allKeysList; el != NULL; el = el->next)
	{
	printf("<A HREF=\"../cgi-bin/hgTracks?position=%s\">", el->name);
	printf("%s </A>", el->name);
	sprintf(query, 
		"select description.name from mrna,description"
		" where mrna.acc = '%s' and mrna.description = description.id",
		el->name);
        if (sqlQuickQuery(conn, query, description, sizeof(description)))
	    printf("- %s", description);
	printf("<BR>\n");
	}
    hFreeConn(&conn);
    }
return TRUE;
}

boolean findMrnaPos(char *acc, char **retChromName, int *retWinStart, int *retWinEnd, boolean *retIsUnique)
/* Look to see if it's an mRNA.  If it aligns uniquely then take them
 * there.  Otherwise take them to a page that lets them choose which
 * alignment. */
{
char *type;
if ((type = mrnaType(acc)) == NULL)
    return FALSE;
else
    {
    struct psl *pslList, *psl;
    int pslCount;
    char prefix[16];

    strncpy(prefix, type, sizeof(prefix));
    tolowers(prefix);
    pslList = psl = findAllAli(acc, prefix);
    pslCount = slCount(pslList);
    if (pslCount <= 0)
        {
	errAbort("%s %s doesn't align anywhere in the draft genome", type, acc);
	}
    else if (pslCount == 1)
        {
	*retChromName = psl->tName;
	*retWinStart = psl->tStart - 100;
	*retWinEnd = psl->tEnd + 100;
	*retIsUnique = TRUE;
	return TRUE;
	}
    else
        {
	slSort(&pslList, pslCmpScore);
	printf("<H2>%s Alignments</H2>", acc);
	printf("This %s aligns in multiple positions.  Click on a hyperlink to ", type);
	printf("go to tracks display at a particular alignment.<BR>");

	printf("<TT><PRE>");
	printf(" SIZE IDENTITY  CHROMOSOME   START     END       cDNA   START  END  TOTAL\n");
	printf("-------------------------------------------------------------------------\n");
	for (psl = pslList; psl != NULL; psl = psl->next)
	    {
	    printf("<A HREF=\"../cgi-bin/hgTracks?position=%s:%d-%d\">",
	        psl->tName, psl->tStart+1, psl->tEnd);
	    printf("%5d  %5.1f%%  %9s  %9d %9d  %8s %5d %5d %5d</A>",
		psl->match + psl->misMatch + psl->repMatch + psl->nCount,
		100.0 - pslCalcMilliBad(psl, TRUE) * 0.1,
		skipChr(psl->tName), psl->tStart + 1, psl->tEnd,
		psl->qName, psl->qStart+1, psl->qEnd, psl->qSize);
	    printf("\n");
	    }
	*retIsUnique = FALSE;
	return TRUE;
	}
    }
}

boolean findGenomePos(char *spec, char **retChromName, 
	int *retWinStart, int *retWinEnd)
/* Find position in genome of spec.  Don't alter
 * return variables if some sort of error. */
{
boolean isUnique = TRUE;
spec = trimSpaces(spec);
if (isHumanChromRange(spec))
    {
    parseHumanChromRange(spec, retChromName, retWinStart, retWinEnd);
    return TRUE;
    }
else if (isContigName(spec))
    {
    findContigPos(spec, retChromName, retWinStart, retWinEnd);
    return TRUE;
    }
else if (findCytoBand(spec, retChromName, retWinStart, retWinEnd))
    {
    return TRUE;
    }
else if (findClonePos(spec, retChromName, retWinStart, retWinEnd))
    {
    return TRUE;
    }
else if (findMrnaPos(spec, retChromName, retWinStart, retWinEnd, &isUnique))
    return isUnique;
else if (findGenethonPos(spec, retChromName, retWinStart, retWinEnd))
    return TRUE;
else if (findMrnaKeys(spec))
    return FALSE;
else
    {
    errAbort("Sorry, couldn't locate %s in genome database\n", spec);
    return TRUE;
    }
}

void doMiddle()
/* Print the body of an html file.  This routine handles zooming and
 * scrolling. */
{
char *oldSeq;
char *submitVal;
boolean testing = FALSE;

// pushCarefulMemHandler(32*1024*1024);
/* Initialize layout and database. */
hSetDb("hg3");
initTl();

/* Read in input from CGI. */
oldSeq = cgiOptionalString("old");
if (oldSeq != NULL)
   calledSelf = TRUE; 
position = cgiOptionalString("position");
if ((submitVal = cgiOptionalString("submit")) != NULL)
    if (sameString(submitVal, "refresh"))
        position = NULL;
if (position == NULL)
    position = "";
if (position != NULL && position[0] != 0)
    {
    if (!findGenomePos(position, &chromName, &winStart, &winEnd))
        return;
    }
if (chromName == NULL)
    {
    chromName = cgiString("seqName");
    winStart = cgiInt("winStart");
    winEnd = cgiInt("winEnd");
    }

/* Clip chromosomal position to fit. */
seqBaseCount = hChromSize(chromName);
seqReverse = cgiBoolean("seqReverse");
if (winEnd < winStart)
    {
    int temp = winEnd;
    winEnd = winStart;
    winStart = temp;
    }
else if (winStart == winEnd)
    {
    winStart -= 1000;
    winEnd += 1000;
    }
if (winStart < 0)
    winStart = 0;
if (winEnd > seqBaseCount)
    winEnd = seqBaseCount;

winBaseCount = winEnd - winStart;
if (winBaseCount == 0)
    errAbort("Window out of range on %s", chromName);

/* Do zoom/scroll if they hit it. */
if (cgiVarExists("left3"))
    relativeScroll(-0.95);
else if (cgiVarExists("left2"))
    relativeScroll(-0.475);
else if (cgiVarExists("left1"))
    relativeScroll(-0.1);
else if (cgiVarExists("right1"))
    relativeScroll(0.1);
else if (cgiVarExists("right2"))
    relativeScroll(0.475);
else if (cgiVarExists("right3"))
    relativeScroll(0.95);
else if (cgiVarExists("in3"))
    zoomAroundCenter(1.0/10.0);
else if (cgiVarExists("in2"))
    zoomAroundCenter(1.0/3.0);
else if (cgiVarExists("in1"))
    zoomAroundCenter(1.0/1.5);
else if (cgiVarExists("out1"))
    zoomAroundCenter(1.5);
else if (cgiVarExists("out2"))
    zoomAroundCenter(3.0);
else if (cgiVarExists("out3"))
    zoomAroundCenter(10.0);

/* Format position string. */
    {
    char buf[256];
    sprintf(buf, "%s:%d-%d", chromName, winStart+1, winEnd);
    position = cloneString(buf);
    }
doForm();
}

void doDown()
{
printf("<H2>The Browser is Being Updated</H2>\n");
printf("The browser is currently unavailable.  We are in the process of\n");
printf("updating the database and the display software with a number of\n");
printf("new tracks, including some gene predictions.  Please try again tomorrow.\n");
}

int main(int argc, char *argv[])
{
htmShell("Draft Genome Browser v2", doMiddle, NULL);
//htmShell("Browser Being Updated", doDown, NULL);
}
